/* wptGPG.h - GnuPG interface
 *	Copyright (C) 2000-2007, 2009 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 */
#ifndef WPT_GPG_H
#define WPT_GPG_H

#include <gpgme.h>
#include "wptKeyCache.h"
#include "wptNLS.h"


#define NO_STRICT 0

/* Macro to indicate if a key is useable or not. */
#define key_is_useable(key) (!(key)->revoked && \
			     !(key)->expired && \
			     !(key)->disabled)

enum setup_t {
    SETUP_KEYGEN    = 1,
    SETUP_IMPORT    = 2,
    SETUP_EXISTING  = 3,
    SETUP_CARDGEN   = 4
};

enum gpg_cmd_t {
    GPG_CMD_DECRYPT = 0,
    GPG_CMD_SIGN    = 1
};

enum keycache_t {
    KEYCACHE_PRV = 0,
    KEYCACHE_PUB = 1
};


struct gpg_card_s;
typedef struct gpg_card_s *gpg_card_t;

/* Keycache refresh context. */
struct refresh_cache_s {
    int kring_update; /* update the keyring cache. */
    int sig_update;   /* update the signature cache. */
};


/* Structure for the passphrase callback. */
struct passphrase_cb_s {
    int gpg_cmd;	/* sign or decrypt. */
    int pwd_init;	/* 1 = passphrase request via dialog. */
    char *pwd;		/* actual passphrase. */
    char info[1024];	/* hold up the info message */
    char *title;	/* dialog title. */
    char keyid[16+1];	/* keyid of the secret key the pwd can be used for. */
    HWND hwnd;		/* handle of the dialog window. */
    int cancel;		/* 1 = user cancelled passphrase request. */
    unsigned int is_card:1; /* 1 = key stored on a smart card. */
    unsigned int bad_pwd:1; /* 1 = last passphrase was bad. */
    gpgme_ctx_t gpg;
    gpgme_recipient_t recipients;
    struct {
	int sym_algo;
	int s2k_mode;
	int s2k_hash;
    } sym;
};


/* This variable is 1 if IDEA is available. */
extern int idea_available;

/*-- wptGPG.cpp --*/
char* get_gnupg_default_key (void);
int set_gnupg_default_key (const char *key);
char* get_gnupg_config (void);
int check_gnupg_config (const char *fname, int *secrings, int *pubrings);
char* get_gnupg_path (void);
int check_gnupg_prog (void);
int gnupg_access_files (void);
char* get_gnupg_prog (void);
char* get_gnupg_keyring (int pub, int strict);
char* get_gnupg_cfgfile (void);
int set_gnupg_options( const char *buf, size_t buflen );
int gnupg_load_config (void);
char* multi_gnupg_path (int strict);

int check_gnupg_options (const char *buf, int showerr);
void init_gnupg_table (void);
void free_gnupg_table (void);
int keyring_check_last_access (void);
const char* gnupg_check_file_ext (const char *fname, int *r_type);
int gpg_check_permissions (int showmsg);
int gpg_check_file_permissions (const char *fname);
int gnupg_check_homedir (void);
int gnupg_access_keyring (int _pub);
void gnupg_backup_options ();
void gnupg_backup_keyrings (int auto_backup, int backup_mode, 
			    int include_secr);
int gnupg_copy_keyrings (void);
int gnupg_import_keypair (void);
int check_gnupg_engine (const char *need_gpg_ver,
			int *r_major, int *r_minor, int *r_patch);

/*-- wptGPGME.cpp --*/
const char * get_signature_status( gpgme_sigsum_t sigstat );

gpgme_error_t get_pubkey (const char *keyid, gpgme_key_t *ret_key);
gpgme_error_t winpt_get_pubkey (const char *keyid, struct winpt_key_s *k);
void	      winpt_release_pubkey (winpt_key_s *k);
gpgme_error_t get_seckey (const char *keyid, gpgme_key_t *ret_skey);
gpgme_error_t winpt_get_seckey (const char *keyid, struct winpt_key_s *k);

void keycache_release (int cleanup);
gpgme_error_t keycache_update (int is_sec, const char *keyid);
gpgme_error_t keycache_init (const char *pubring, const char * secring);
gpg_keycache_t keycache_get_ctx (int _pub);

gpgme_error_t gpg_clip_sym_encrypt (void);
const char *get_gpg_sigstat (gpgme_sigsum_t sum);
int check_ultimate_trusted_key (void);
bool secret_key_available (void);

/*-- wptGPGParser.cpp --*/
enum entry_t {
    ENTRY_OPAQUE = 1,
    ENTRY_SINGLE = 2,
    ENTRY_MULTI  = 3,
    ENTRY_GROUP  = 4,
};

struct conf_option_s {
    struct conf_option_s *next;
    char *name;
    char *val;
    unsigned int used:1;
    int type;  
};
typedef struct conf_option_s *conf_option_t;

struct conf_member_s {
    struct conf_member_s * next;
    char *name;
    unsigned int used:1;  
};
typedef struct conf_member_s *conf_member_t;

struct conf_group_s {
    struct conf_group_s *next;
    struct conf_member_s *list;
    char *name;
    unsigned int used:1;  
};
typedef struct conf_group_s *conf_group_t;

struct conf_file_s {
    conf_option_t list;
    conf_group_t grp;
};
typedef struct conf_file_s *config_file_t;

/* find/delete/add */
conf_group_t conf_find_group (config_file_t opt, const char *str);
conf_option_t conf_find_option (config_file_t opt, const char *str);
conf_member_t conf_find_member (config_file_t opt, 
				const char *grp, const char *str);
int conf_delete_group (config_file_t opt, const char *str);
int conf_delete_member (config_file_t opt, const char *grp, const char *str);
int conf_delete_option (config_file_t opt, const char *str);
int conf_modify_entry (config_file_t opt, int type, const char *name, const char *val);
int conf_add_entry (config_file_t opt, int type, const char *name, const char *val);
int conf_add_entry_int (config_file_t opt, int type, const char *name, int val);
int conf_add_member (config_file_t opt, const char *grp, const char *str);
int conf_add_group (config_file_t opt, const char *str);
/* high-level */
void new_config (config_file_t *r_opt);
int parse_config (const char *file, config_file_t *r_opt);
int commit_config (const char *file, config_file_t opt);
void release_config (config_file_t opt);
void release_group (conf_group_t grp);

/*-- wptPassphraseCB.cpp --*/
enum passdlg_t {
    PASSDLG_REPEAT  = 0, /* indicate the dialig is in the 'repeat passwd' state. */
    PASSDLG_INIT    = 1,
    PASSDLG_STRICT  = 2,
    PASSDLG_NOTEMPTY= 4,
    PASSDLG_WARN_UTF8=8,
};

void set_gpg_passphrase_cb (passphrase_cb_s *cb, gpgme_ctx_t ctx,
			    int cmd, HWND hwnd, const char *title);
void release_gpg_passphrase_cb (passphrase_cb_s *cb);
void release_gpg_recipients (gpgme_recipient_t *recipients);

char *request_passphrase (const char *title, int flags, int *ret_cancel);
char *request_passphrase2 (const char *title, int flags, int *ret_cancel);
char *request_key_passphrase (gpgme_key_t key, const char *title, 
			      int *ret_cancel);
//const char * passphrase_cb (void *opaque, const char * desc, void *r_hd);
char * get_key_userid (const char *keyid);
int check_passwd_quality (const char *pass, int strict);

/*-- wptClipboard.cpp --*/
gpgme_error_t gpg_clip_istext_avail (int *r_val);
gpgme_error_t gpg_clip_is_secured (int *r_type, int *r_val);
gpgme_error_t gpg_clip_get_pgptype (int *r_type);
gpgme_error_t gpg_clip_parse_pgpid (const char *data, int *r_type);

gpgme_error_t gpg_data_utf8_new_from_clipboard (gpgme_data_t *r_dh, 
						int wraplen, int *r_is_utf8);
gpgme_error_t gpg_data_new_from_clipboard (gpgme_data_t *r_dh, 
					   int wraplen);

gpgme_error_t gpg_data_release_utf8_to_clipboard (gpgme_data_t dh);
gpgme_error_t gpg_data_release_to_clipboard (gpgme_data_t dh, int chg_ver);
						  

/* wptGPGMEData.cpp --*/

size_t       gpg_data_readline (gpgme_data_t dh, char *line, size_t nbytes);

gpgme_error_t gpg_data_release_and_set_file (gpgme_data_t dh, const char *fname);
gpgme_error_t  gpg_data_mail_quote (gpgme_data_t *r_dh);
gpgme_error_t  gpg_data_extract_plaintext (gpgme_data_t sig, gpgme_data_t *r_plain);
void	      gpg_data_putc (gpgme_data_t hd, int c);

/*-- wptGPGUtil.cpp --*/
gpgme_error_t gpg_rebuild_cache (char **r_inf);
gpgme_error_t gpg_get_version (char **r_inf);
void          gpg_set_debug_mode (int val);
gpgme_error_t gpg_export_seckey (const char *keyid, const char *outfile);
gpgme_error_t gpg_manage_ownertrust (char **data, int do_export);
gpgme_error_t gpg_get_photoid_data (const char *keyid, char **r_status_data,
				    unsigned char **r_data, 
				    unsigned long *ndata);
gpgme_error_t gpg_revoke_cert (int desig_revoke, const char *inp_data, 
			       const char *keyid, char **r_revcert);
gpgme_error_t gpg_decode_c_string (const char *src, char **destp, size_t len);
gpgme_error_t gpg_import_key_list (const char *fname, char **r_out);
gpgme_error_t gpg_extract_keys (const char *keyfile, const char **keys, DWORD nkeys,
				char **new_keyfile);
gpgme_error_t get_uat_validity (const char *keyid, gpgme_validity_t *r_valid);
gpgme_error_t gpg_get_recipients (const char *file, gpgme_recipient_t *r_list);
gpgme_error_t gpg_find_key_subpacket (const char *key, int subpktid,
				      char **value);
int gpg_check_return_code (const char *args);

#endif /* WPT_GPG_H */
