#ifndef __STRING_BUFFER_H
#define __STRING_BUFFER_H

/* Copyright (C) 2006, 2007, 2009 Timo Schulz
   Released under the GNU General Public License, Version 2. */

#define DEFSIZE 128

/* A simple string buffer implementation which allows to manipulate
   the underlying buffer in a unified and easy way. The code is written
   for clarity and simplicty, _not_ speed. */
class StringBuffer
{
private:
    char *private_contents;
    int alloced_size;
    int size;

private:
    /* Add the converted data at the end of the internal buffer.
       Allocate more memory if needed, plus some extra bytes. */
    void addCharacters (const char *buffer, int bufsize);

public:
    StringBuffer ();

    /* Create an empty buffer object with an inital size. */
    StringBuffer (int bufsize);


    /* Create a buffer object with the given contents. */
    StringBuffer (const char *contents);

    /* Destructor which frees the used memory. */
    ~StringBuffer ();

    /* Overwrite buffer contents in a safe manner. */
    void wipeContents (void);

    /* Return a constant pointer to the 'char*' data of the buffer. */
    const char *getBuffer (void);

    /* Return a copy of the buffer which must be freed after use. */
    char* getBufferCopy (void);

    /* Return the size of the underlying string. */
    int getSize (void);

    /* Reset the buffer contents. */
    void reset (void);

    /* Add an integer to the buffer. */
    void add (int val);

    /* Add a single character to the buffer. */
    void add (char val);

    /* Add a long-integer to the buffer. */
    void add (long val);

    /* Add a string to the buffer. */
    void add (const char *val);
    
    /* Add only @len bytes of @val. */
    void add(const char *val, size_t len);

    /* Add an unsigned integer to the buffer. */
    void add (unsigned long val);

    /* Add an unsigned integer in hex format to the buffer. */
    void addHex (unsigned long val);

    /* Overloaded operators to provide an easier interface
       for manipulating the underlying buffer. */
    StringBuffer& operator +(int val);

    StringBuffer& operator +(unsigned long val);

    StringBuffer& operator +(const char *val);

    StringBuffer& operator +=(int val);

    StringBuffer& operator +=(unsigned long val);

    StringBuffer& operator +=(const char *val);
    
    StringBuffer& operator +=(char *);

    StringBuffer& operator +=(char val);

    StringBuffer& operator=(const char *val);
    
    StringBuffer& operator=(char *val);
};


#endif /*__STRING_BUFFER_H*/
