/* wptBalloonPop.cpp - Balloon Popup-Box
 *	Copyright (C) 2009 Timo Schulz
 * 
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 */


/* The idea is to use a balloon style 'error message' which
   points to the control the error happened at. For instance
   if no valid E-Mail address was entered, the balloon message
   contains an error description with an arrow to the edit field
   the address was supposed to be entered in. */

#include <windows.h>
#include <stdio.h>
#include <time.h>

#include "wptTypes.h"

#define BT_WINDOWNAME		"WinPT_BalloonPop"
#define BT_TIMER_IDENTIFIER	23
#define BT_IOFFSET               8


extern HINSTANCE glob_hinst;

struct balloon_ctx_s {
    LPCTSTR text;
    LPCTSTR icon_name;
    POINT   pos;
    HWND    hwnd;
};
static struct balloon_ctx_s glob_msg;
static BOOL window_class_registered = FALSE;


static void 
display_balloon_text (HWND hwnd, HDC hdc, const char *text)
{
    RECT rc;    
    memset (&rc, 0, sizeof (rc));    
    GetClientRect(hwnd, &rc);
    
    POINT pts[3];
    pts[0].x = rc.left + BT_IOFFSET;
    pts[0].y = rc.top;
    pts[1].x = pts[0].x;
    pts[1].y = pts[0].y + BT_IOFFSET;
    pts[2].x = pts[1].x + BT_IOFFSET;
    pts[2].y = pts[1].y;
    
    // we merge three regions here:
    // 1) basis
    // 2) rounded box
    // 3) arrow
    HRGN hrgn = CreateRectRgn(0, 0, 0, 0);    
    HRGN hrgn1 = CreateRoundRectRgn(rc.left, 
				    rc.top + BT_IOFFSET,
				    rc.right,
				    rc.bottom,
				    15, 15);
    HRGN hrgn2 = CreatePolygonRgn(&pts[0], 3, ALTERNATE);
    CombineRgn(hrgn, hrgn1, hrgn2, RGN_OR);
    
    // fill the region and draw a frame  around it
    FillRgn (hdc, hrgn, GetSysColorBrush(COLOR_INFOBK));
    FrameRgn (hdc, hrgn, (HBRUSH)GetStockObject (DKGRAY_BRUSH), 1, 1);
    
    rc.top = rc.top + BT_IOFFSET*2;
    rc.bottom = rc.bottom - BT_IOFFSET;
    rc.left = rc.left + BT_IOFFSET;
    rc.right = rc.right - BT_IOFFSET;
    
    // FIXME: calculate best position
    HICON hicon = LoadIcon (NULL, glob_msg.icon_name);
    if (hicon != NULL) {
	DrawIconEx (hdc, 5, 20, hicon, 28, 28, 0, NULL, DI_NORMAL);
	DestroyIcon (hicon);
    }
    
    SetTextColor (hdc, GetSysColor (COLOR_INFOTEXT));
    DrawText (hdc, text, strlen (text), &rc, DT_VCENTER + DT_NOCLIP);
}


// window procedure for the balloon tip.
static LRESULT CALLBACK
window_proc (HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    switch (msg) {
    case WM_PAINT:
	RECT rc;
	GetWindowRect (hwnd, &rc);
	
	POINT curpos;
	curpos.x = glob_msg.pos.x;
	curpos.y = glob_msg.pos.y;
	
	rc.right = curpos.x - BT_IOFFSET + 6 + rc.right - rc.left;
	rc.bottom = curpos.y + 20 + rc.bottom - rc.top;
	rc.left = curpos.x - BT_IOFFSET + 6;
	rc.top = curpos.y + 20;
	MoveWindow (hwnd, rc.left, rc.top, rc.right - rc.left,
		    rc.bottom - rc.top, FALSE);
	
	HDC hdc;
	PAINTSTRUCT ps;
	hdc = BeginPaint (hwnd, &ps);
	
	LOGFONT lf;
	memset (&lf, 0, sizeof (lf));
	lf.lfHeight = 13;
	lf.lfWeight = FW_NORMAL;
	lf.lfQuality = ANTIALIASED_QUALITY;
	strcpy (lf.lfFaceName, "MS Sans Serif");
	
	HFONT hfont;
	hfont = CreateFontIndirect(&lf);
	HFONT hfont_old;
	hfont_old = (HFONT)SelectObject(ps.hdc, hfont);
	
	DrawText (ps.hdc, glob_msg.text, strlen (glob_msg.text),
		  &rc, DT_VCENTER+DT_NOCLIP+DT_CALCRECT);
	rc.right = rc.right + 2*BT_IOFFSET;
	rc.bottom = rc.bottom + 3*BT_IOFFSET;
	
	ShowWindow (hwnd, SW_SHOWNA);
	MoveWindow(hwnd, rc.left, rc.top, rc.right - rc.left,
		   rc.bottom - rc.top, TRUE);		   
	SetBkMode(ps.hdc, TRANSPARENT);
	display_balloon_text (hwnd, ps.hdc, glob_msg.text);
	
	// restore old font object
	SelectObject(ps.hdc, hfont_old);
	DeleteObject (hfont);
	EndPaint (hwnd, &ps);
	break;

    case WM_LBUTTONUP:
	ShowWindow (hwnd, SW_HIDE);
	break;

    default:
	return DefWindowProc (hwnd, msg, wparam, lparam);
    }

    return FALSE;
}


/* Callback procedure to hide the baloon window after the
   time has been expired */
static VOID CALLBACK 
popup_timer_proc (HWND hwnd, UINT uMsg, UINT_PTR idEvent, DWORD dwTime)
{
    ShowWindow (hwnd, SW_HIDE);
}


/* Register our new window class */
static void
register_window_class (HINSTANCE hinst)
{
    WNDCLASS wclass;    
    memset (&wclass, 0, sizeof (wclass));
    wclass.style = CS_HREDRAW | CS_VREDRAW;
    wclass.lpfnWndProc = window_proc;
    wclass.hInstance = hinst;
    wclass.lpszClassName = BT_WINDOWNAME;
    
    if (!RegisterClass (&wclass))
	abort ();
}


static LPCTSTR
format_text (LPCTSTR string)
{
    const int off = 12;    
    static char buf[2048];    
    int n = strlen (string), pos=0, i;
    
    memset (buf, 0, DIM (buf));
    // FIXME: rewrite the function
    if (n > 1600)
	abort ();

    buf[pos++] = '\n';
    for (i=0; i < off; i++)
	buf[pos++] = ' ';
    for (i=0; i < n; i++) {
	buf[pos++] = string[i];
	if (string[i] == '\n') {
	    for (int j=0; j < off; j++)
		buf[pos++] = ' ';
	}
    }
    buf[pos++] = '\n';
    return buf;
}


/* Display a baloon message box at the given X- Y-coordinates */
HWND
show_balloon_msg_pos (HWND hparwnd, int millis, int x, int y,
		      LPCTSTR string, LPCTSTR icon_name)
{
    if (!window_class_registered) {
	register_window_class (glob_hinst);
	window_class_registered = TRUE;
    }

    // if no icon is requsted, print the text as it is.
    if (!icon_name)
	glob_msg.text = string;
    else
	glob_msg.text = format_text (string);
    glob_msg.icon_name = icon_name;
    
    if (x > 0 && y > 0) { // use specific coordinates
	glob_msg.pos.x = x;
	glob_msg.pos.y = y;
    }
    else { // point to the lower left region of the control
	RECT rect;
	GetWindowRect (hparwnd, &rect);
	glob_msg.pos.x = rect.left;
	glob_msg.pos.y = rect.bottom - 20;
    }
    
    // we use a persistent window for the balloon messages
    // instead of destroy/create we hide the window whenever it
    // is unused.
    if (!glob_msg.hwnd) {
	glob_msg.hwnd = CreateWindowEx (WS_EX_TOOLWINDOW | WS_EX_TOPMOST,
					BT_WINDOWNAME, BT_WINDOWNAME,
					WS_POPUP, 1, 1, 1, 1,
					NULL, NULL, glob_hinst, NULL);
	if (!glob_msg.hwnd)
	    BUG (0);
    }
    
    // ensure 'parent' never loses focus.
    HWND hwnd = glob_msg.hwnd;
    ShowWindow (hwnd, SW_SHOW);
    UpdateWindow (hwnd);
    EnableWindow (hparwnd, TRUE);
    SetActiveWindow (NULL);

    // show message for a fixed amount of time
    SetTimer (hwnd, BT_TIMER_IDENTIFIER, millis, popup_timer_proc);

    return hwnd;
}


/* Show message at the default position for 5 seconds */
void
show_balloon_msg (HWND hparwnd, LPCTSTR string, LPCTSTR icon_name)
{
    show_balloon_msg_pos (hparwnd, 5000, 0, 0, string ,icon_name);
}


/* Disable the message box (hide the window) */
void
balloon_msg_disable (void)
{
    ShowWindow (glob_msg.hwnd, SW_HIDE);
}
