/* wptCardEditCB.cpp - Card callbacks
 *	Copyright (C) 2003, 2004, 2005, 2006 Timo Schulz
 *	Copyright (C) 2005 g10 Code GmbH
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, gpg_write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "gpgme.h"
#include "wptCard.h"
#include "wptCardEdit.h"
#include "wptErrors.h"

#define save_write(val) ((val)? (val) : "")
static const char *stat_key = "";


/* Wrapper to emulate write(). */
static int
gpg_write (int fd, const void *buf, size_t buflen)
{
    HANDLE hd = (HANDLE)fd;
    DWORD n;

    /*
    #ifdef _DEBUG
    MessageBox (NULL, (const char*)buf, stat_key, MB_OK);
    #endif
    */
    if (!WriteFile (hd, buf, buflen, &n, NULL))
	log_debug ("card gpg_write failed ec=%d\r\n", (int)GetLastError ());
    if (!WriteFile (hd, "\n", 1, &n, NULL))
	log_debug ("card gpg_write failed ec=%d\r\n", (int)GetLastError ());
    return n;
}


/* Common card handler. Handle general card status events. */
static gpgme_error_t
common_card_handler (GpgCardEdit *c, gpgme_status_code_t code, 
		     const char *key, int fd)
{
    const char *s;

    if (!strcmp (key,  "cardctrl.insert_card.okay")) {
	if (c->card_cb) {
	    s = c->card_cb (CARD_CTL_INSERT, c->cb_value);
	    gpg_write (fd, s, strlen (s));
	    return 0;
	}
    }

    if (code == GPGME_STATUS_CARDCTRL && 
	(*key == CARD_CTL_NO_CARD || *key == CARD_CTL_NO_READER))
	c->setResult (GPG_CARDRES_NOCARD);

    /* This can happen for example if a wrong PIN were
       submitted. In such a case, we quit ASAP. */
    if (code == GPGME_STATUS_SC_OP_FAILURE) {
	if (key && !strcmp (key, "2"))
	    c->setResult (GPG_CARDRES_BAD_PIN);
	else if (key && !strcmp (key, "1"))
	    c->setResult (GPG_CARDRES_CANCEL);
	s = "quit\n";
	gpg_write (fd, s, strlen (s));
	/* XXX: do we need c->reset()? */
	return 0;
    }
    return 0;
}


/* handler for changing PINs of cards. */
static gpgme_error_t
change_pin_handler (void *opaque, gpgme_status_code_t code, 
		    const char *key, int fd)
{
    GpgCardEdit *c = (GpgCardEdit *)opaque;
    const char *s = "";
    int type;

    if (!c)
	return gpg_error (GPG_ERR_EOF);

    stat_key = key; /* XXX: debug code */

    common_card_handler (c, code, key, fd);

    type = c->getType ();
    if (!strcmp (key,  "cardedit.prompt")) {
	if (c->cnt == 0) {
	    c->cnt++;
	    s = "admin";
	}
	else if (c->cnt == 1) {
	    c->cnt++;
	    s = "passwd";
	}
	else {
	    c->reset ();
	    s = "quit";
	}
	gpg_write (fd, s, strlen (s));
	return 0;
    }
    if (!strcmp (key,  "cardutil.change_pin.menu")) {
	if (c->cnt == 2) {
	    c->cnt++;
	    switch (type) {
	    case GPG_EDITCARD_CHUPIN: s = "1"; break;
	    case GPG_EDITCARD_UNBPIN: s = "2"; break;
	    case GPG_EDITCARD_CHAPIN: s = "3"; break;
	    default:		      s = "Q";
	    }
	    gpg_write (fd, s, strlen (s));
	    return 0;
	}
	else if (c->cnt > 0) {
	    gpg_write (fd, "Q", 1);
	    return 0;
	}
    }
    if (type == GPG_EDITCARD_CHUPIN) {
	if (!strcmp (key, "passphrase.pin.ask")) {
	    s = c->pin;
	    c->cnt++;
	}
	if (!strcmp (key, "passphrase.pin.new.ask")) {
	    s = c->pin_new;
	    c->cnt++;
	}
	if (!strcmp (key, "passphrase.pin.repeat") || 
	    !strcmp (key, "passphrase.ask")) {
	    s = c->pin_new;
	    c->cnt++;
	}
	gpg_write (fd, s, strlen (s));
    }
    else if (type == GPG_EDITCARD_CHAPIN) {
	if (!strcmp (key, "passphrase.adminpin.ask")) {
	    s = c->admin_pin;
	    c->cnt++;
	}
	if (!strcmp (key, "passphrase.adminpin.new.ask")) {
	    s = c->pin_new;
	    c->cnt++;
	}
	if (!strcmp (key, "passphrase.pin.repeat") ||
	    !strcmp (key, "passphrase.ask")) {
	    s = c->pin_new;
	    c->cnt++;
	}
	gpg_write (fd, s, strlen (s));
    }
    else if (type == GPG_EDITCARD_UNBPIN) {
	/* todo */
    }

    return 0;
}


/* edit card attributes handler. */
static gpgme_error_t
editcard_handler (void * opaque, gpgme_status_code_t code, 
		  const char *key, int fd)
{
    GpgCardEdit *c = (GpgCardEdit *)opaque;   
    const char *s = "";
    int type;

    if (!c)
	return gpg_error (GPG_ERR_EOF);

    stat_key = key;

    common_card_handler (c, code, key, fd);

    type = c->getType ();
    if (!strcmp (key, "cardedit.prompt")) {
	if (c->getResult ()) { /* in case of an error we quit. */
	    c->cnt = 0;
	    s = "quit";
	    gpg_write (fd, s, strlen (s));
	    return 0;
	}
	if (c->cnt == 0) { /* first switch in the admin mode */
	    c->cnt++;
	    s = "admin";
	    gpg_write (fd, s, strlen (s));
	    return 0;
	}
	if (c->cnt == 1) {/* then run the send command */
	    c->cnt++;
	    switch (type) {
	    case GPG_EDITCARD_NAME:   s = "name"; break;
	    case GPG_EDITCARD_KEYURL: s = "url"; break;
	    case GPG_EDITCARD_LOGIN:  s = "login"; break;
	    case GPG_EDITCARD_SEX:    s = "sex"; break;
	    case GPG_EDITCARD_LANG:   s = "lang"; break;
	    case GPG_EDITCARD_GENKEY: s = "generate"; break;
	    default:		      s =  "quit";
	    }
	    gpg_write (fd, s, strlen (s));
	    return 0;
	}
	else if (c->cnt >= 2) {/* done: send exit */
	    c->reset ();
	    s = "quit";
	    gpg_write (fd, s, strlen (s));
	    return 0;
	}
    }
    if (c->cnt > 0) {
	if (!strcmp (key, "passphrase.adminpin.ask")) {
	    gpg_write (fd, c->admin_pin, strlen (c->admin_pin));
	    return 0;
	}
	if (!strcmp (key, "passphrase.pin.ask")) {
	    gpg_write (fd, c->pin, strlen (c->pin));
	    return 0;
	}
    }
    switch (type) {
    case GPG_EDITCARD_NAME:
	if (!strcmp (key, "keygen.smartcard.surname")) {
	    s = save_write (c->edit.surname);
	    gpg_write (fd, s, strlen (s));
	}
	else if (!strcmp (key, "keygen.smartcard.givenname")) {
	    s = save_write (c->edit.givenname);
	    gpg_write (fd, s, strlen (s));
	}
	break;

    case GPG_EDITCARD_KEYURL:
	if (!strcmp (key, "cardedit.change_url")) {
	    s = save_write (c->edit.keyurl);
	    gpg_write (fd, s, strlen (s));
	}
	break;

    case GPG_EDITCARD_LOGIN:
	if (!strcmp (key, "cardedit.change_login"))  {
	    s = save_write (c->edit.login);
	    gpg_write (fd, s, strlen (s));
	}
	break;

    case GPG_EDITCARD_SEX:
	if (!strcmp (key, "cardedit.change_sex")) {
	    static char buf[2];
	    if (c->edit.sex != 'M'
		&& c->edit.sex != 'F'
		&& c->edit.sex != ' ') {
		buf[0] = ' '; 
		buf[1] = 0;
	    }
	    else {
		buf[0] = c->edit.sex; 
		buf[1] = 0;
	    }
	    gpg_write (fd, buf, strlen (buf));
	}
	break;

    case GPG_EDITCARD_LANG:
	if (!strcmp (key, "cardedit.change_lang")) {
	    s = save_write (c->edit.lang);
	    gpg_write (fd, s, strlen (s));
	}
	break;

    case GPG_EDITCARD_GENKEY:
	if (!strcmp (key, "passphrase.enter"))
	    gpg_write (fd, c->keygen.pass, strlen (c->keygen.pass));
	if (!strcmp (key, "cardedit.genkeys.backup_enc")) {
	    s = c->keygen.flags & GPG_CARDFLAG_BAKENC? "Y" : "N";
	    gpg_write (fd, s, strlen (s));
	}
	if (!strcmp (key, "cardedit.genkeys.replace_keys")) {
	    if (! (c->keygen.flags & GPG_CARDFLAG_REPLACE))
		c->cancel = 1;
	    s = c->keygen.flags & GPG_CARDFLAG_REPLACE? "Y" : "N";
	    gpg_write (fd, s, strlen (s));
	}
	else if (!strcmp (key, "keygen.valid")) {
	    s = save_write (c->keygen.expdate);
	    gpg_write (fd, s, strlen (s));
	}
	else if (!strcmp (key, "keygen.name")) {
	    s = save_write (c->keygen.name);
	    gpg_write (fd, s, strlen (s));
	}
	else if (!strcmp (key, "keygen.email")) {
	    s = save_write (c->keygen.email);
	    gpg_write (fd, s, strlen (s));
	}
	else if (!strcmp (key, "keygen.comment")) {
	    s = save_write (c->keygen.comment);
	    gpg_write (fd, s, strlen (s));
	}
	break;

    default:
	break;
    }
    
    return 0;
}


/* Check if a GPG status code occured which marks the
   current operation as failed.
   Return value: gpg error constant. */
static gpgme_error_t
map_result (GpgCardEdit *ce)
{
    if (!ce->getResult ())
	return gpg_error (GPG_ERR_NO_ERROR);
    if (ce->getResult () & GPG_CARDRES_NOCARD)
	return gpg_error (GPG_ERR_CARD_NOT_PRESENT);
    if (ce->getResult () & GPG_CARDRES_BAD_PIN)
	return gpg_error (GPG_ERR_BAD_PIN);
    return 0;
}


/* Wrapper around the gpgme card interface.
   @ctx: context to use.
   @ce:  card edit context.
   Return value: 0 on success. */
gpgme_error_t
gpg_card_edit (gpgme_ctx_t ctx, GpgCardEdit *ce)
{
    gpgme_data_t out;
    gpgme_error_t err;

    err = gpgme_data_new (&out);
    if (err)
	return err;

    if (ce->getType () > GPG_EDITCARD_CHPIN_ID)
	err = gpgme_op_card_edit (ctx, NULL, change_pin_handler, ce, out);
    else
	err = gpgme_op_card_edit (ctx, NULL, editcard_handler, ce, out);
    if (!err)
	err = map_result (ce);
    
    gpgme_data_release (out);
    return err;
}
