/* wptClipEncryptDlg.cpp - Clipboard encrypt dialog
 *	Copyright (C) 2000-2006, 2009 Timo Schulz
 *	Copyright (C) 2005 g10 Code GmbH
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <commctrl.h>

#include "resource.h"
#include "wptTypes.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptErrors.h"
#include "wptCommonCtl.h"
#include "wptGPG.h"
#include "wptKeylist.h"
#include "wptNLS.h"
#include "wptContext.h" /* for passphrase_s */
#include "wptRegistry.h"
#include "wptDlgs.h"
#include "wptUTF8.h"
#include "StringBuffer.h"


/* Encrypt the contents of the clipboard with the keys in @rset.
   If @always_trust is set, GPG is forced to trust any recipients.
   The context is returned in @r_ctx.
   Return value: 0 on success. */
gpgme_error_t
gpg_clip_encrypt (gpgme_key_t rset[], int always_trust, gpgme_ctx_t *r_ctx)
{
    gpgme_error_t err;
    gpgme_encrypt_flags_t flags;
    gpgme_ctx_t ctx = NULL;
    gpgme_data_t plain = NULL;
    gpgme_data_t ciph = NULL;
    
    err = gpgme_new (&ctx);
    if (err)
        return err;

    /* Recently a lot of users got in trouble because they enabled
       the textmode even for non-text (binary) data. This leads to
       serious trouble. But because our clipboard functions only support
       the text format, the returned data is pure text or at least a 
       conversion of it, so we can safely use the textmode here. */
    gpgme_set_textmode (ctx, 1);
    gpgme_set_armor (ctx, 1);
    
    err = gpg_data_utf8_new_from_clipboard (&plain, 0, NULL);
    if (err)
	goto leave;
    err = gpgme_data_new (&ciph);
    if (err)
	goto leave;
    flags = always_trust? GPGME_ENCRYPT_ALWAYS_TRUST : 
      (gpgme_encrypt_flags_t)0;
    err = gpgme_op_encrypt (ctx, rset, flags, plain, ciph);
    *r_ctx = ctx;
    if (err)
	goto leave;
    
    gpg_data_release_to_clipboard (ciph, 1);
    ciph = NULL;

leave:
    if (ciph)
	gpgme_data_release (ciph);
    gpgme_data_release (plain);
    return err;
}


/* Show all invalid recipients if possible. @dlg is the 
   handle to the calling dialog and @ctx is the context
   used for encryption.
   Return value: 0 if invalid recipients exist -1 otherwise. */
static int
show_invalid_recipients (HWND dlg, gpgme_ctx_t ctx)
{
    gpgme_encrypt_result_t res;
    gpgme_invalid_key_t k;
    gpgme_key_t key;
    StringBuffer p;
    const char *keyid;
    const char *warn = _("key not found");
    char *uid;

    if (!ctx)
	return -1;
    res = gpgme_op_encrypt_result (ctx);
    if (!res || !res->invalid_recipients)
	return -1;

    p = _("Recipients unusable for encryption:\n");
    for (k = res->invalid_recipients; k; k = k->next) {
	if (!get_pubkey (k->fpr, &key)) {
	    uid = utf8_to_native (key->uids->name);
	    keyid = key->subkeys->keyid+8;
	}
	else {
	    uid = strdup (warn);
	    keyid = k->fpr;
	}
	p = p + keyid + " : " + uid + "\n";
	safe_free (uid);
    }
    msg_box (dlg, p.getBuffer (), _("Encryption"), MB_ERR);
    return 0;
}


struct dlg_ctx_s {
    int keys_sortby;
    int enable_redraw;
    keylist_ctrl_t kl;
};
// FIXME: associate the context with the window
static struct dlg_ctx_s state;

/* Dialog procedure for the clipboard encryption. */
BOOL CALLBACK 
clip_encrypt_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{    
    gpg_keycache_t kc;
    gpgme_key_t *rset;
    gpgme_ctx_t ctx=NULL;
    gpgme_error_t err;
    char tmpbuf[64];
    size_t n;    
    
    switch (msg) {
    case WM_INITDIALOG:
        SetWindowText (dlg, _("Encryption"));
	SetDlgItemText (dlg, IDC_ENCRYPT_FINDKEY, _("&Find Key:"));
	SetDlgItemText (dlg, IDCANCEL, _("&Cancel"));
	SetDlgItemText (dlg, IDC_ENCRYPT_FNDCMD, _("&Find")); 
        kc = keycache_get_ctx (KEYCACHE_PUB);
        state.kl = keylist_load (GetDlgItem (dlg, IDC_ENCRYPT_KEYLIST), kc, NULL,
				 KEYLIST_ENCRYPT_MIN, KEY_SORT_USERID);	

	center_window (dlg, NULL);
        SetForegroundWindow (dlg);

	state.enable_redraw = 1;
	InvalidateRect(GetDlgItem (dlg, IDC_ENCRYPT_KEYLIST), NULL, TRUE);
        return TRUE;
        
    case WM_DESTROY:
        if (state.kl) {
            keylist_delete (state.kl);
            state.kl = NULL;
        }
        break;
        
    case WM_NOTIFY:
	NMHDR *notify;        
        notify = (NMHDR *)lparam;
	if (!notify)
	    break;

	/*
	if (state.enable_redraw) {
	    int ret;
	    ret = keylist_listview_notify (dlg, state.kl->keys,
					    IDC_ENCRYPT_KEYLIST, lparam);
	    if (ret != 0) {
		SetWindowLong (dlg, DWL_MSGRESULT, ret);
		return TRUE;
	    }
	}*/
	
        if (notify && notify->code == NM_DBLCLK && 
	    notify->idFrom == IDC_ENCRYPT_KEYLIST)
            PostMessage( dlg, WM_COMMAND, MAKEWPARAM (IDOK, 0), 0);
	if (notify && notify->code == LVN_COLUMNCLICK && 
	    notify->idFrom == IDC_ENCRYPT_KEYLIST) {
	    NMLISTVIEW *p = (LPNMLISTVIEW) lparam;
	    int sortby = 0;

	    switch (p->iSubItem) {
	    case  0: sortby = KEY_SORT_USERID; break;
	    case  1: sortby = KEY_SORT_KEYID; break;
	    case  2: sortby = KEY_SORT_LEN; break;
	    case  4: sortby = KEY_SORT_VALIDITY; break;
	    default: sortby = KEY_SORT_USERID; break;
	    }
	    if ((state.keys_sortby & ~KEYLIST_SORT_DESC) == sortby)
		state.keys_sortby ^= KEYLIST_SORT_DESC;
	    else
		state.keys_sortby = sortby;
	    keylist_sort (state.kl, state.keys_sortby);
	}
        break;
        
    case WM_COMMAND:
	state.enable_redraw = 0;

        switch (LOWORD (wparam)) {
        case IDOK:
	    int force_trust;
            rset = keylist_get_recipients (state.kl, &force_trust, &n);
	    if (!n) {
                msg_box (dlg, _("You must select at least one key."), 
			 _("Encryption"), MB_ERR);
                safe_free (rset);
                return FALSE;
            }
	    err = gpg_clip_encrypt (rset, force_trust, &ctx);
            if (err) {
		if (show_invalid_recipients (dlg, ctx))
		    msg_box (dlg, gpgme_strerror (err), _("Encryption"), MB_ERR);
            }
            else
		show_msg (dlg, 1500, _("GnuPG Status: Finished"));
            safe_free (rset);
	    if (ctx)
		gpgme_release (ctx);
	    if (!err)
		EndDialog (dlg, TRUE);
            break;
            
        case IDCANCEL:
            EndDialog (dlg, FALSE);
            break;

	case IDC_ENCRYPT_FNDCMD:
	    n = GetDlgItemText (dlg, IDC_ENCRYPT_FIND, tmpbuf, DIM (tmpbuf)-1);
	    if (!n)
		break;
	    n = listview_find (state.kl->lv, tmpbuf, 0);
	    if (n != -1) {
		int oldpos = listview_get_curr_pos (state.kl->lv);
		listview_select_one (state.kl->lv, n);
		listview_scroll (state.kl->lv, oldpos, n);
	    }
	    else
		log_box (_("Encryption"), MB_ERR, 
			 _("No recipient found with '%s'"), tmpbuf);
	    break;
        }
        break;
    }
    
    return FALSE;
}

