/* wptFileStatDlg.cpp - List the status of an OpenPGP file
 *	Copyright (C) 2002-2003, 2005-2006, 2009 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <time.h>
#include <sys/types.h>
#include <assert.h>

#include "resource.h"
#include "wptTypes.h"
#include "wptGPG.h"
#include "wptCommonCtl.h"
#include "wptContext.h"
#include "wptErrors.h"
#include "wptW32API.h"
#include "wptUTF8.h"
#include "openpgp.h"


/* Make sure that the list won't consume too much memory. */
#define MAX_PKTS 1024

/* Return humand readable public key algo. */
static const char*
get_pubkey_algo (int algo)
{	
    if (algo == 17)
	return "DSA";	
   else if (is_ELGAMAL (algo))
	return "ELG";	
    else if (is_RSA (algo))
	return "RSA";
    return NULL;
}


/* Return human readable symmetric algo. */
const char*
get_symkey_algo (int algo)
{
    switch (algo) {
    case 0: return "PLAINTEXT";
    case 1: return "IDEA";
    case 2: return "3DES"; 
    case 3: return "CAST5";
    case 4: return "BLOWFISH";
    case 7: return "AES128";
    case 8: return "AES192";
    case 9: return "AES256";
    default: break;
    }    
    return "UNKNOWN";
}


/* Return human readable hash algo. */
static const char*
get_digest_algo (int algo)
{
    switch (algo) {
    case 0: return "NONE";
    case 1: return "MD5";
    case 2: return "SHA1";
    case 3: return "RMD160";
    case 5: return "MD2";
    case 8: return "SHA256";
    case 9: return "SHA384";
    case 10:return "SHA512";
    case 11:return "SHA224";
    default: break;
    }
    return "UNKNOWN";
}


/* Return human readable compress algo. */
static const char*
get_compress_algo (int algo)
{
    switch (algo) {
    case 0: return "NONE";
    case 1: return "ZIB";
    case 2: return "ZLIB";
    case 3: return "BZIP2";
    default: break;
    }
    return "UNKNOWN";
}


/* Return human string to key algo. */
static const char*
get_s2k_algo (int algo)
{
    switch (algo) {
    case 0: return "SIMPLE";
    case 1: return "SALTED";
    case 3: return "ITER+SALTED";
    default:break;
    }
    return "INVALID";
}


/* List all OpenPGP packets from the file @file.
   Return value: number of packets. */
static int
do_list_packets (HWND list, const char *file)
{
    PACKET *pkt;
    gpg_iobuf_t inp = NULL;
    armor_filter_context_t afx;
    char inf[1024], *p;
    char fmtdate[128];
    size_t n;
    int rc;    

    if (!list || !file) {
	log_debug ("do_list_packets: invalid parameters\n");
	return -1;
    }

    inp = gpg_iobuf_open (file);
    if (!inp) {
	log_box (_("File Manager"), MB_ERR, "%s: %s", file,
		 winpt_strerror (WPTERR_FILE_OPEN));
	return -1;
    }
    gpg_iobuf_ioctl (inp, 3, 1, NULL); /* disable cache */
    if (gpg_use_armor_filter (inp)) {
	memset (&afx, 0, sizeof (afx));
	gpg_iobuf_push_filter (inp, gpg_armor_filter, &afx);
    }
    pkt = (PACKET *)calloc(1, sizeof *pkt);
    gpg_init_packet (pkt);
    n=0;
    while ((rc = gpg_parse_packet (inp, pkt)) != -1) {
	switch (pkt->pkttype) {
	case PKT_PUBKEY_ENC:
	    {PKT_pubkey_enc *enc = pkt->pkt.pubkey_enc; n++;
	    if (!enc)
		break;
	    _snprintf (inf, DIM(inf)-1,
		       "public key encryted packet: "
		       "version %d, algo %s, keyid 0x%08X",
		       enc->version, get_pubkey_algo(enc->pubkey_algo), 
		       enc->keyid[1]);
	    listbox_add_string (list, inf);}
	    break;
		
	case PKT_SYMKEY_ENC:
	    {PKT_symkey_enc *enc = pkt->pkt.symkey_enc;	n++;
		if (!enc)
		break;
	    _snprintf (inf, DIM(inf)-1,
		       "symmetric key encrypted packet: "
		       "version %d, cipher %s, s2k %s, hash %s",
		       enc->version, get_symkey_algo (enc->cipher_algo),
		       get_s2k_algo (enc->s2k.mode), 
		       get_digest_algo (enc->s2k.hash_algo));
	    listbox_add_string (list, inf); }
	    break;

	case PKT_ENCRYPTED:	
	case PKT_ENCRYPTED_MDC:
	    {PKT_encrypted *enc = pkt->pkt.encrypted; n++;
	    if (!enc)
		break;
	    _snprintf (inf, DIM (inf)-1,	
		       "encrypted data packet: mdc method %d, length %d",
		       enc->mdc_method, enc->len);
	    listbox_add_string (list, inf); }
	    break;

	case PKT_PUBLIC_KEY:
	case PKT_PUBLIC_SUBKEY:
	    {PKT_public_key *pk = pkt->pkt.public_key; n++;
	    if (!pk)
		break;
	    _snprintf (inf, DIM(inf)-1,
		       "public key packet: version %d, algo %s, created %s",
		       pk->version, get_pubkey_algo (pk->pubkey_algo),
		       get_locale_date (pk->timestamp, fmtdate, 
					DIM (fmtdate)-1));
	    listbox_add_string (list, inf); }
	    break;

	case PKT_SECRET_KEY:
	case PKT_SECRET_SUBKEY:	    
	    {PKT_secret_key *sk = pkt->pkt.secret_key; n++;
	    if (!sk)
		break;
	    _snprintf (inf, DIM (inf)-1,
		       "secret key packet: "
		       "version %d, algo %s, created %s mode %d",
		       sk->version,  get_pubkey_algo (sk->pubkey_algo),
		       get_locale_date (sk->timestamp, fmtdate, 
					DIM(fmtdate)-1),
		       sk->protect.s2k.mode);
	    listbox_add_string (list, inf); }
	    break;
	    
	case PKT_SIGNATURE:
	    {PKT_signature *sig = pkt->pkt.signature; n++;
	    if (!sig)
		break;
	    _snprintf (inf, DIM(inf)-1,
		       "signature packet: "
		       "version %d, algo %s, keyid 0x%08X, created %s",
		       sig->version, get_pubkey_algo(sig->pubkey_algo), 
		       sig->keyid[1],
		       get_locale_date (sig->timestamp, fmtdate, 
					DIM(fmtdate)-1));
	    listbox_add_string (list, inf); }
	    break;

		
	case PKT_USER_ID:
	    {PKT_user_id *id = pkt->pkt.user_id; n++;
	    if (!id)
		break;
	    p = utf8_to_native (id->name);
	    _snprintf (inf, DIM (inf)-1, "user id packet: %s", p);
	    free_if_alloc (p);
	    listbox_add_string (list, inf); }
	    break;
	    
	case PKT_COMPRESSED:
	    {PKT_compressed *zip = pkt->pkt.compressed; n++;
	    if (!zip)
		break;
	    _snprintf (inf, DIM (inf)-1, 
		       "compressed packet: algo %s, length %d",
		       get_compress_algo(zip->algorithm), zip->len);
	    listbox_add_string (list, inf); }
	    break;

	default:
	    break;
	}
	gpg_free_packet (pkt);
	gpg_init_packet (pkt);
	if (n > MAX_PKTS) {
	    listbox_add_string(list, "--last shown packet--");
	    break;
	}
    }
    gpg_iobuf_close (inp);
    safe_free (pkt);
    return n;
}


/* Dialog box procedure for showing file status. */
BOOL CALLBACK
file_stat_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    switch (msg) {
    case WM_INITDIALOG:
	if (do_list_packets (GetDlgItem (dlg, IDC_FILE_STAT_LIST), 
			     (const char*)lparam) < 1) {
	    msg_box (dlg, _("No valid OpenPGP data found."), 
		     _("File Status"), MB_ERR);
	    EndDialog (dlg, FALSE);
	    return TRUE;
	}
	SetWindowText (dlg, _("File Status"));
        SetForegroundWindow (dlg);
        break;
        
    case WM_COMMAND:
        switch (LOWORD (wparam)) {
        case IDOK:
            EndDialog (dlg, TRUE);
            break;

	case IDCANCEL:
	    EndDialog (dlg, FALSE);
	    break;
        }
        break;
    }
    
    return FALSE;
}
