/* wptJPG.cpp - Routines for showing JPG pictures
 *      Copyright (C) 2005-2006, 2008-2009, 2011 Timo Schulz
 *	Copyright (C) 2001 Dr.Yovav Gad
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/*-----------------------------------------------------------------------------
 * Picture (Implementations) Version 1.00
 * Author: Dr. Yovav Gad, EMail: Sources@SuperMain.com
 * Web: www.SuperMain.com
 * 
 * This version uses a stripped down and heavily modified version of 
 * Picture.cpp and Picture.h.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <ocidl.h>
#include <olectl.h>

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>

#include "wptJPG.h"
#include "wptErrors.h"
#include "wptTypes.h"

#define HIMETRIC_INCH	2540
#define ERROR_TITLE	"WinPT - CJPG Error"


/* Constructor to create an empty JPG container. */
CJPG::CJPG (void)
{
    m_IPicture = NULL;
    m_Height = 0;
    m_Weight = 0;
    m_Width = 0;    
}


/* Deconstructor. Free all internal data. */
CJPG::~CJPG (void)
{
    if (m_IPicture != NULL)
	freePictureData (); 
}



/* Free the allocated memory that holdes the IPicture Interface data
   and clear picture information. */
void 
CJPG::freePictureData (void)
{
    if (m_IPicture != NULL) {
	m_IPicture->Release();
	m_IPicture = NULL;
    }
    m_Height = 0;
    m_Weight = 0;
    m_Width = 0;
}


/* Open a JPG File And Load It Into IPicture (Interface) */
BOOL 
CJPG::load (LPCSTR filepath)
{    
    FILE *fp;

    if (m_IPicture != NULL) 
	freePictureData ();

    fp = fopen (filepath, "rb");
    if (!fp) {
	MessageBox (NULL, strerror (errno), ERROR_TITLE, MB_OK|MB_ICONSTOP);
	return FALSE;
    }

    /* avoid to load empty JPG files and make sure we still can
       access the file handle. */
    struct stat st;
    if (fstat (fileno (fp), &st) || st.st_size == 0) {
	fclose (fp);
	return FALSE;
    }
    BYTE *buffer = new BYTE[st.st_size];
    if  (!buffer)
	BUG (0);
    memset (buffer, 0, st.st_size);
    int n = fread (buffer, 1, st.st_size, fp);
    fclose (fp);

    /* not the entire file were read in, so abort here. */
    if (n != st.st_size) {
	delete []buffer;
	return FALSE;
    }
    if (!loadPictureData (buffer, st.st_size)) {
	delete []buffer;
	return FALSE;
    }
    delete [] buffer;

    m_Weight = st.st_size; /* Update Picture Size Info... */
    if (m_IPicture == NULL) {
	m_Height = 0;
	m_Width = 0;
	return FALSE;
    }
    m_IPicture->get_Height (&m_Height);
    m_IPicture->get_Width (&m_Width);
    /* Calculate Its Size On a "Standard" (96 DPI) Device Context */
    m_Height = MulDiv (m_Height, 96, HIMETRIC_INCH);
    m_Width  = MulDiv (m_Width,  96, HIMETRIC_INCH);	
    return TRUE;
}



/* read the picture data from a source (file / resource)
   and load it into the current IPicture object in use */
BOOL 
CJPG::loadPictureData (BYTE *buffer, int nsize)
{
    BOOL result = FALSE;
    
    HGLOBAL hglobal;
    hglobal = GlobalAlloc (GMEM_MOVEABLE, nsize);
    if (hglobal == NULL)
	BUG (0);

    void *data;
    data = GlobalLock (hglobal);
    memcpy (data, buffer, nsize);
    GlobalUnlock (hglobal);
    
    IStream *stream = NULL;
    if (CreateStreamOnHGlobal (hglobal, TRUE, &stream) != S_OK) {
	GlobalFree (hglobal);
	return result;
    }	
	
    HRESULT hr;
    hr = OleLoadPicture (stream, nsize, FALSE, IID_IPicture,
			 (LPVOID *)&m_IPicture);
    if (hr == E_NOINTERFACE) {
	MessageBox (NULL, "IPicture interface is not supported",
		    ERROR_TITLE, MB_OK|MB_ICONSTOP);
    }
    else if (hr == E_POINTER) {
	MessageBox (NULL, "The provided stream is not valid",
		    ERROR_TITLE, MB_ERR);
    }
    else { /* S_OK */
	stream->Release ();
	stream = NULL;
	result = TRUE;
    }

    GlobalFree (hglobal);
    return result;
}


/* Draw the loaded picture direct to the client DC */
BOOL 
CJPG::show (HDC pDC, POINT *leftTop, POINT *widthHeight, 
	    int magnifyX, int magnifyY)

{
    RECT drawRect;
    HRESULT hr;
    long width  = 0;
    long height = 0;

    if (pDC == NULL || m_IPicture == NULL) 
	return FALSE;

    m_IPicture->get_Width (&width);
    m_IPicture->get_Height (&height);
	
    if (magnifyX == 0) 
	magnifyX = 0;
    if (magnifyY == 0) 
	magnifyY = 0;
    magnifyX = int(MulDiv (width, GetDeviceCaps(pDC, LOGPIXELSX), HIMETRIC_INCH) * magnifyX);
    magnifyY = int(MulDiv (height, GetDeviceCaps(pDC, LOGPIXELSY), HIMETRIC_INCH) * magnifyY);

    drawRect.left = leftTop->x;
    drawRect.top = leftTop->y;
    drawRect.right = magnifyX;
    drawRect.bottom = magnifyY;

    hr = m_IPicture->Render (pDC,
			    leftTop->x,               // Left
			    leftTop->y,               // Top
			    widthHeight->x +magnifyX, // Width
			    widthHeight->y +magnifyY, // Height
			    0,
			    height,
			    width,
			    -height,
			    &drawRect);
    if (SUCCEEDED (hr)) 
	return TRUE;
    return FALSE;
}


/* Get the original picture pixel size (ignore what current DC is using)
   pointer to a Device Context is needed for pixel calculation, */
BOOL 
CJPG::updateSizeOnDC (HDC pDC)

{	
    if(pDC == NULL || m_IPicture == NULL) { 
	m_Height = 0; 
	m_Width = 0; 
	return FALSE;
    }

    m_IPicture->get_Height (&m_Height);
    m_IPicture->get_Width (&m_Width);

    /* Get Current DPI - Dot Per Inch */
    int CurrentDPI_X = GetDeviceCaps (pDC, LOGPIXELSX);
    int CurrentDPI_Y = GetDeviceCaps (pDC, LOGPIXELSY);

    m_Height = MulDiv (m_Height, CurrentDPI_Y, HIMETRIC_INCH);
    m_Width  = MulDiv (m_Width,  CurrentDPI_X, HIMETRIC_INCH);

    return TRUE;
}


/* Return height of the current image. */
LONG
CJPG::getHeight (void)
{
    return m_Height;
}


/* Return weight of the current image. */
LONG 
CJPG::getWeight (void)
{
    return m_Weight;
}


/* Return width of the current image. */
LONG
CJPG::getWidth (void)
{
    return m_Width;
}

/* Display a JPG picture in the given window at the given point. */
int
jpg_show (HWND hwnd, POINT *p, LPCSTR name)
{
    CJPG jpg;        
    BOOL rc;

    rc = jpg.load (name);
    if (!rc)
	return WPTERR_GENERAL;
    HDC hdc = GetWindowDC (hwnd);
    rc = jpg.updateSizeOnDC (hdc);
    if (!rc) {
	ReleaseDC (hwnd, hdc);
	return WPTERR_GENERAL;
    }

    POINT sizewnd;
    RECT rwnd;
    GetWindowRect (hwnd, &rwnd);    
    sizewnd.x = rwnd.right - rwnd.left;
    sizewnd.y = rwnd.bottom - rwnd.top;
    rc = jpg.show (hdc, p, &sizewnd, 0, 0);

    ReleaseDC (hwnd, hdc);
    jpg.freePictureData ();
    return rc == TRUE? WPTERR_SUCESS : WPTERR_GENERAL;
}
