/* wptKeyEdit.cpp - GPG key edit abstraction
 *	Copyright (C) 2005, 2006 Timo Schulz
 *	Copyright (C) 2005 g10 Code GmbH
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <time.h>

#include "gpgme.h"
#include "wptCommonCtl.h"
#include "wptContext.h"
#include "wptKeyEdit.h"
#include "wptTypes.h"
#include "wptW32API.h"
#include "wptGPG.h"
#include "wptErrors.h"
#include "wptUTF8.h"


/* Parse the colon status information of @line and store
   the information in @rev.
   Return value: 0 on success. */
static gpgme_error_t
rev_key_colon_handler (gpg_desig_rev_t *rev, char *line)
{
    char *p, *pend;
    gpg_desig_rev_t r, t;
    int field = 0;

    if (!line || strlen (line) < 3)
	return gpg_error (GPG_ERR_EOF);
    if (strncmp (line, "rvk", 3))
	return 0; /* skip this line. */

    log_debug ("rev_key: line=%s\r\n", line);

    r = (gpg_desig_rev_t)calloc (1, sizeof *r);
    if (!r)
	return gpg_error (GPG_ERR_ENOMEM);
    if (!*rev)
	*rev = r;
    else {
	for (t=*rev; t->next; t=t->next)
	    ;
	t->next = r;
    }

    p = strdup (line);
    if (!p)
	return gpg_error (GPG_ERR_ENOMEM);

    for (;;) {
	field++;
	pend = strsep (&p, ":");
	if (pend == NULL)
	    break;
	switch (field) {
	case  4: 
	    r->pubkey_algo = (gpgme_pubkey_algo_t)atol (pend); 
	    break;

	case 10: 
	    strncpy (r->fpr, pend, 40); 
	    r->fpr[40] = 0; 
	    break;
	}	
    }
    free (p);
    return 0;
}


/* Parse the colon data output of edit key from @line and
   store the information in the @inf context.
   Return value: 0 on success. */
static gpgme_error_t
uid_inf_colon_handler (gpg_uid_info_t *inf, char *line)
{
    gpg_uid_info_t i, t;
    char *p, *pend;
    char *name;
    int field = 0, len = 0;

    if (!line || strlen (line) < 3 || strncmp (line, "uid", 3))
        return gpg_error (GPG_ERR_EOF);

    i = (gpg_uid_info_t)calloc (1, sizeof *i);
    if (!i)
	return gpg_error (GPG_ERR_ENOMEM);
    if (!*inf)
	*inf = i;
    else {
	for (t=*inf; t->next; t=t->next)
	    ;
	t->next = i;
    }

    p = strdup (line);
    if (!p)
	return gpg_error (GPG_ERR_ENOMEM);;
    for (;;) {
        field++;
        pend = strsep (&p, ":");
        if (pend == NULL)
            break;

        switch (field) {
        case 2: /* trust info */
            break;

        case 10: /* user ID */
            name = (char *)calloc (1, strlen (pend)+1);
	    if (!name)
		return gpg_error (GPG_ERR_ENOMEM);;
	    gpg_decode_c_string (pend, &name, strlen (pend)+ 1);
	    i->name = utf8_to_native (name);
	    safe_free (name);
	    if (strchr (pend, '<') != NULL && strchr (pend, '>') != NULL) {
		int pos = strchr (i->name, '<')- i->name + 1;
		int end = strchr (i->name, '>') - i->name;
		i->email = (char*) calloc (1, end-pos+2);
		if (!i->email)
		    return gpg_error (GPG_ERR_ENOMEM);;
		memcpy (i->email, i->name+pos, (end-pos));
	    }
            break;

        case 13: /* preferences */
            if (strstr (pend, "mdc")) {
		const char *s = "no-ks-modify";
                len = strlen (pend) - 4; /* ,mdc */
		if (strstr (pend, s)) {
		    i->flags.no_ks_modify = 1;
		    len -= strlen (s)+1; /* ',' + s */
		}
                i->prefs = (char*)calloc (1, len+1);
		if (!i->prefs)
		    return gpg_error (GPG_ERR_ENOMEM);
                memcpy (i->prefs, pend, len);
                i->prefs[len] = '\0';
                i->flags.mdc = 1;
            }
            else {
                i->prefs = strdup (pend);
		if (!i->prefs)
		    return gpg_error (GPG_ERR_ENOMEM);
                i->flags.mdc = 0;		
            }
            break;

        case 14: /* index/flags */
            i->index = atoi (pend);
	    if (strchr (pend, 'r'))
		i->flags.revoked = 1;
	    if (strchr (pend, 'p'))
		i->flags.primary = 1;
            break;
        }
    }
    free (p);
    return 0;
}


/* Release the context in @inf. */
void
gpg_uid_info_release (gpg_uid_info_t list)
{
    gpg_uid_info_t i;
   
    while (list) {
	i = list->next;
	if (list->name) {
	    if (list->name)	     
		free (list->name);
	    list->name = NULL;
	}
	if (list->prefs) {
	    if (list->prefs)
		free (list->prefs);
	    list->prefs = NULL;
	}	
	free (list);
	list = i;
    }
}


/* Release the context in @rev. */
void
gpg_desig_rev_release (gpg_desig_rev_t rev)
{
    gpg_desig_rev_t r;

    while (rev) {
	r = rev->next;
	free (rev);
	rev = r;
    }
}


static gpgme_error_t
list2_handler (void *opaque, gpgme_status_code_t code, const char *key, int fd)
{
    DWORD n;
    const char *s;

    if (!strcmp (key, "keyedit.prompt")) {
	s = "quit\n";
	WriteFile ((HANDLE)fd, s, strlen (s), &n, NULL);
    }    
    return 0;
}


/* Dummy handler to get the colon data and then quit. */
static gpgme_error_t
list_handler (void *opaque, gpgme_status_code_t code, const char *key, int fd)
{
    static int step = 0;
    const char *s = "";
    DWORD n;

    if (!strcmp (key, "keyedit.prompt") && step == 0) {
	step = 1;
	s = "list\n";
	WriteFile ((HANDLE)fd, s, strlen (s), &n, NULL);
    }
    else if (!strcmp (key, "keyedit.prompt") && step == 1) {
	step = 0;
	s = "quit\n";
	WriteFile ((HANDLE)fd, s, strlen (s), &n, NULL);
    }

    return 0;
}


/* Return all designated revokers for this key. If no revoker
   was set, @r_rev is NULL.
   Return value: 0 on success. */
gpgme_error_t
GpgKeyEdit::getDesignatedRevoker (gpg_desig_rev_t *r_rev)
{
    gpgme_data_t out = NULL;
    gpg_desig_rev_t rev = NULL;
    gpgme_error_t err;
    char buf[256];

    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);

    err = gpgme_data_new (&out);
    if (err)
	goto leave;
    err = gpgme_op_edit (ctx, key, list2_handler, NULL, out);
    if (err)
	goto leave;
    
    gpgme_data_rewind (out);
    while (gpg_data_readline (out, buf, DIM (buf)-1) > 0)
	rev_key_colon_handler (&rev, buf);
    *r_rev = rev;

leave:
    if (out)
	gpgme_data_release (out);
    if (err) {
	gpg_desig_rev_release (rev);
	*r_rev = NULL;
    }
    return err;
}


/* Retrieve all user ID information of the key set via setKey 
   in @r_inf. The result also contains the user ID number which 
   is needed to securely delete the user-ID. */
gpgme_error_t
GpgKeyEdit::getUseridInfo (gpg_uid_info_t *r_uinf)
{
    gpgme_data_t out = NULL;    
    gpgme_error_t err;
    gpg_uid_info_t inf = NULL;
    char buf[256];

    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);

    err = gpgme_data_new (&out);
    if (err)
	goto leave;
    err = gpgme_op_edit (ctx, key, list_handler, NULL, out);
    if (err)
	goto leave;
    
    gpgme_data_rewind (out);
    while (gpg_data_readline (out, buf, DIM (buf) -1) > 0)
	uid_inf_colon_handler (&inf, buf);

    *r_uinf = inf;

leave:
    if (out)
	gpgme_data_release (out);
    if (err) {
	gpg_uid_info_release (inf);
	*r_uinf = NULL;
    }
    return err;
}


/* Clear object. */
void
GpgKeyEdit::clear (void)
{
    pass = NULL;
    name = NULL;
    cmt = NULL;
    email = NULL;
    type = 0;
    cnt = 0;
    cmd_sent = 0;
    resval = 0;
    uid_index = -1;
    sig_index = -1;
    key_index = -1;
    key_has_passwd = true;
}


GpgKeyEdit::GpgKeyEdit (void)
{    
    clear ();
    key = NULL;
    gpgme_new (&ctx); /*FIXME */
    
}


/* Construct an object with the given key in @key. */
GpgKeyEdit::GpgKeyEdit (gpgme_key_t _key)
{
    clear ();
    this->key = _key;
    gpgme_new (&ctx); /* FIXME */
}

/* Construct an object and fetch the key with the keyid @keyid. */
GpgKeyEdit::GpgKeyEdit (const char *_keyid)
{
    clear ();
    get_pubkey (_keyid, &this->key); /*FIXME: check return code*/
    gpgme_new (&ctx); /* FIXME */
}

/* Delete the given object. */
GpgKeyEdit::~GpgKeyEdit (void)
{
    free_if_alloc (name);
    free_if_alloc (cmt);
    free_if_alloc (email);    
    gpgme_release (ctx);
}


/* Reset the state of the object. */
void
GpgKeyEdit::reset (void)
{
    cmd_sent = 0;
    cnt = 0;
}


/* Return true if type has a non-zero value. */
bool
GpgKeyEdit::isValid (void)
{
    return type != 0;
}


/* Return the GPGME key. */
gpgme_key_t
GpgKeyEdit::getKey (void)
{
    return key;
}


/* Set the GPGME callback to @cb. The hook value can be
   given in @cb_value. */
void 
GpgKeyEdit::setCallback (gpgme_progress_cb_t cb, void *cb_value)
{
    gpgme_set_progress_cb (ctx, cb, cb_value);
}


/* Clear the stored passphrase. */
void
GpgKeyEdit::clearPassphrase (void)
{
    if (pass)
	pass = NULL;
}



/* Inidicate that a key is protected by  a passphrase or not. */
void
GpgKeyEdit::setNoPassphrase (bool val)
{
    key_has_passwd = !val;
}


void
GpgKeyEdit::setPassphraseCallback (gpgme_passphrase_cb_t cb, void *cb_value)
{    
    key_has_passwd = false;
    gpgme_set_passphrase_cb (ctx, cb, cb_value);
}


bool
GpgKeyEdit::usePassphraseCallback (void)
{
    gpgme_passphrase_cb_t cb_ptr;
    void *cb_value;

    gpgme_get_passphrase_cb (ctx, &cb_ptr, &cb_value);
    if (cb_ptr)
	return true;
    return false;
}

/* Set the passphrase to @pass. */
void 
GpgKeyEdit::setPassphrase (const char *_pass)
{
    this->pass = _pass;
}


/* Kludge to allow to handle admin and user PINs. */
void
GpgKeyEdit::setPassphrase2 (const char *_pass2)
{
    this->new_pass = _pass2;
}

/* Set the current key to @key. */
void 
GpgKeyEdit::setKey (gpgme_key_t _key)
{
    this->key = _key;
}

/* Set the keyid of the destination key to @keyid. */
void
GpgKeyEdit::setKeyID (const char *_keyid)
{
    if (!_keyid)
	return;
    get_pubkey (_keyid, &this->key); /* FIXME: check return code */
}


/* Set the local user for the operation to @locusr. */
void 
GpgKeyEdit::setLocalUser (gpgme_key_t locusr)
{
    gpgme_signers_add (ctx, locusr);
}

/* Set the result of the operation to @val. */
void 
GpgKeyEdit::setResult (int val)
{
    resval |= val;
}


/* Return the result of the operation. */
int
GpgKeyEdit::getResult(void)
{
    return resval;
}


/* Return the amount of days the key is valid. */
int
GpgKeyEdit::getValidDays (void)
{
    return valid;
}


int
GpgKeyEdit::getType (void)
{
    return type;
}


/* Add the notation data from @notation to the user ID
   with the index @_uid_idx.
   Return value: 0 on success.  */
gpgme_error_t 
GpgKeyEdit::addNotation (int _uid_idx, const char *_notation)
{
    if (!key)
	return gpg_error (GPG_ERR_INV_OBJ);
    if (key_has_passwd && !this->pass)
	return gpg_error (GPG_ERR_INV_PASSPHRASE);

    type = GPG_EDITKEY_NOTATION;
    this->uid_index = _uid_idx;
    this->notation = (char*)_notation;
    return gpg_editkey (this->ctx, this->key, this);
}


/* Sign the key stored in the object with the 
   signing mode @mode and the signature class @sig_class.
   Return value: 0 on success. */
gpgme_error_t
GpgKeyEdit::signKey (int mode, int _sig_class, const char *_exp_date)
{
    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);
    if (key_has_passwd && !this->pass)
	return gpg_error (GPG_ERR_INV_PASSPHRASE);

    type = mode;
    this->exp_date = _exp_date;
    this->sig_class = _sig_class;
    return gpg_editkey (this->ctx, this->key, this);
}


/* Sign a single user-id with the index @_uid_index.
   All other parameters are equal to signKey().
   Return value: 0 on success. */
gpgme_error_t
GpgKeyEdit::signUserid (int _uid_idx, int mode, int _sig_class, 
			const char *_exp_date)
{
    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);
    if (key_has_passwd && !this->pass)
	return gpg_error (GPG_ERR_INV_PASSPHRASE);

    this->uid_index = _uid_idx;
    type = mode;
    this->exp_date = _exp_date;
    this->sig_class = _sig_class;
    return gpg_editkey (this->ctx, this->key, this);
}


/* Set the ownertrust of the key stored in the object
   to the trust value @trust.
   Return value: 0 on success. */
gpgme_error_t
GpgKeyEdit::setTrust (gpgme_validity_t trust)
{
    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);

    type = GPG_EDITKEY_TRUST;
    this->trust_id = (int)trust;
    return gpg_editkey (this->ctx, this->key, this);
}

/* Add a user ID to the given key with the @name as the
   name, @cmt as the comment (or NULL) and @email as the email.
   Return value: 0 on success. */
gpgme_error_t
GpgKeyEdit::addUserid (const char *_name, const char *_cmt, const char *_email)
{
    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);
    if (key_has_passwd && !this->pass)
	return gpg_error (GPG_ERR_INV_PASSPHRASE);

    type = GPG_EDITKEY_ADDUID;
    free_if_alloc (this->name);
    this->name = m_strdup (_name);
    free_if_alloc (this->cmt);
    this->cmt = NULL;
    if (_cmt != NULL)
	this->cmt = m_strdup (_cmt);	
    free_if_alloc (this->email);
    this->email = m_strdup (_email);
    if (!this->email || !this->name)
	BUG (NULL);
    return gpg_editkey (this->ctx, this->key, this);
}

/* Delete the user-ID with the index @uid_index of the given key.
   Return value: 0 on success. */
gpgme_error_t
GpgKeyEdit::delUserid (int _uid_index)
{
    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);

    type = GPG_EDITKEY_DELUID;
    this->uid_index = _uid_index;
    return gpg_editkey (this->ctx, this->key, this);
}

/* Delete the subkey with the index @key_index.
   Return value: 0 on success. */
gpgme_error_t
GpgKeyEdit::delKey (int _key_index)
{
    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);

    type = GPG_EDITKEY_DELKEY;
    this->key_index = _key_index;
    return gpg_editkey (this->ctx, this->key, this);
}


/* Add a new key on a smart card. 
   The type is given in @type. */
gpgme_error_t 
GpgKeyEdit::addCardKey (int _type)
{
    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);
    if (!this->pass)
	return gpg_error (GPG_ERR_INV_PASSPHRASE);
    if (_type < 1 || _type > 3)
	return gpg_error (GPG_ERR_INV_VALUE);

    type = GPG_EDITKEY_ADDCARDKEY;
    this->key_index = _type;
    return gpg_editkey (this->ctx, this->key, this);
}


/* Add a new subkey to the given key.
   The new key will have @pubkey_algo as the algorithm
   and a size of @pubkey_size bits. If valid > 0, the
   key expires in @valid days. 
   Return value: 0 on success. */
gpgme_error_t
GpgKeyEdit::addSubkey (gpgme_pubkey_algo_t _pubkey_algo, 
		       unsigned int _pubkey_size, long _valid)
{
    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);
    if (key_has_passwd && !this->pass)
	return gpg_error (GPG_ERR_INV_PASSPHRASE);

    type = GPG_EDITKEY_ADDKEY;
    this->pubkey_algo = _pubkey_algo;
    this->pubkey_size = _pubkey_size;
    this->valid = _valid;
    return gpg_editkey (this->ctx, this->key, this);
}

/* Change the passphrase of the given key to @new_pass.
   If allow_empty != 0, it is allowed to provide an empty passphrase.
   Return value: 0 on success. */
gpgme_error_t
GpgKeyEdit::changePassphrase (const char *_new_pass, int allow_empty)
{
    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);
    if (key_has_passwd && !this->pass)
	return gpg_error (GPG_ERR_INV_PASSPHRASE);
    
    type = GPG_EDITKEY_PASSWD;
    this->new_pass = _new_pass;
    this->flags = allow_empty? 1 : 0;
    return gpg_editkey (this->ctx, this->key, this);
}

/* Set the primary user-ID of the given key to user-ID with
   the index @uid_index.
   Return value: 0 on success. */
gpgme_error_t
GpgKeyEdit::setPrimaryUserid (int _uid_index)
{
    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);
    if (key_has_passwd && !this->pass)
	return gpg_error (GPG_ERR_INV_PASSPHRASE);

    type = GPG_EDITKEY_PRIMARY;
    this->uid_index = _uid_index;
    return gpg_editkey (this->ctx, this->key, this);
}

/* Set the expire date of the subkey with the index @key_index.
   @exp_timestamp is used to calculate the days the key is valid.
   if @exp_days is true, exp_timestamp is already converted to days.
   Return value: 0 on success. */
gpgme_error_t
GpgKeyEdit::setKeyExpireDate (int _key_index, long exp_timestamp, 
			      bool exp_days)
{
    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);
    if (key_has_passwd && !this->pass)
	return gpg_error (GPG_ERR_INV_PASSPHRASE);
    if (!exp_days && exp_timestamp > 0 && exp_timestamp < time (NULL))
	return gpg_error (GPG_ERR_INV_ARG);
    
    type = GPG_EDITKEY_EXPIRE;
    if (!exp_days && exp_timestamp > 0) {
	valid = exp_timestamp - time (NULL);
	valid /= 86400;
    }
    else
	valid = exp_timestamp;
    this->key_index = _key_index;
    return gpg_editkey (this->ctx, this->key, this);
}

/* Revoke the userid given by the index @uid_index.
   Return value: 0 on success. */
gpgme_error_t
GpgKeyEdit::revokeUserid (int _uid_index)
{
    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);
    if (key_has_passwd && !this->pass)
	return gpg_error (GPG_ERR_INV_PASSPHRASE);

    type = GPG_EDITKEY_REVUID;
    this->uid_index = _uid_index;
    return gpg_editkey (this->ctx, this->key, this);
}


/* Revoke a signature on the user-ID with the index @uid_index
   and the signature index @sig_index.
   Return value: 0 on success. */
gpgme_error_t
GpgKeyEdit::revokeSignature (int _uid_index, int _sig_index)
{
    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);
    if (key_has_passwd && !this->pass)
	return gpg_error (GPG_ERR_INV_PASSPHRASE);

    type = GPG_EDITKEY_REVSIG;
    this->uid_index = _uid_index;
    this->sig_index = _sig_index;
    return gpg_editkey (this->ctx, this->key, this);
}


/* Revoke the subkey with the index @key_index. Optionally
   a reason can be provided in @reason with a text to describe
   more details in @cmt.
   Return value: 0 on success. */
gpgme_error_t
GpgKeyEdit::revokeSubkey (int _key_index, int _reason, const char *_cmt)
{
    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);
    if (key_has_passwd && !this->pass)
	return gpg_error (GPG_ERR_INV_PASSPHRASE);

    type = GPG_EDITKEY_REVKEY;
    this->key_index = _key_index;
    this->reason = _reason;
    free_if_alloc (this->cmt);
    this->cmt = NULL;
    if (_cmt)
	this->cmt = m_strdup (_cmt);
    return gpg_editkey (this->ctx, this->key, this);
}


/* Add a designated revoker to the key. @uid stores
   the user-ID of the key who is allowed to be a
   designated revoker.
   Return value: 0 on success. */
gpgme_error_t
GpgKeyEdit::addDesignatedRevoker (const char *uid)
{
    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);
    if (key_has_passwd && !this->pass)
	return gpg_error (GPG_ERR_INV_PASSPHRASE);

    type = GPG_EDITKEY_ADDREV;
    free_if_alloc (this->name);
    this->name = m_strdup (uid);
    return gpg_editkey (this->ctx, this->key, this);
}

/* Add a photo-ID to the key. The JPG image is given
   in the file with the name @jpg_file.
   Return value: 0 on success. */
gpgme_error_t
GpgKeyEdit::addPhotoid (const char *jpg_file)
{
    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);
    if (key_has_passwd && !this->pass)
	return gpg_error (GPG_ERR_INV_PASSPHRASE);

    type = GPG_EDITKEY_ADDPHOTO;
    this->url = jpg_file;
    return gpg_editkey (this->ctx, this->key, this);
}

/* Enable the given key. */
gpgme_error_t
GpgKeyEdit::enable (void)
{
    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);
    type = GPG_EDITKEY_ENABLE;
    return gpg_editkey (this->ctx, this->key, this);
}

/* Disable the given key. */
gpgme_error_t
GpgKeyEdit::disable (void)
{
    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);
    type = GPG_EDITKEY_DISABLE;
    return gpg_editkey (this->ctx, this->key, this);
}


/* Remove unusable parts and all signatures from a key. */
gpgme_error_t 
GpgKeyEdit::minimizeKey (void)
{
    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);
    type = GPG_EDITKEY_MINIMIZE;
    return gpg_editkey (this->ctx, this->key, this);
}


/* Remove unusable parts from a key. */
gpgme_error_t
GpgKeyEdit::cleanKey (void)
{
    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);
    type = GPG_EDITKEY_CLEAN;
    return gpg_editkey (this->ctx, this->key, this);
}


/* Update the user-ID preferences of the user-ID with the
   index @uid_index to the prefs given in @new_prefs.
   Return value: 0 on success. */
gpgme_error_t
GpgKeyEdit::setUseridPreferences (int _uid_index, const char *_new_prefs)
{
    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);
    if (key_has_passwd && !this->pass)
	return gpg_error (GPG_ERR_INV_PASSPHRASE);
    type = GPG_EDITKEY_SETPREF;
    this->uid_index = _uid_index;
    this->new_prefs = _new_prefs;
    return gpg_editkey (this->ctx, this->key, this);
}


/* Delete a signature from the user-ID with the index @uid_index.
   The index of the signature is given in @sig_index.
   Return value: 0 on success. */
gpgme_error_t
GpgKeyEdit::delUseridSignature (int _uid_index, int _sig_index)
{
    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);
    type = GPG_EDITKEY_DELSIG;
    this->uid_index = _uid_index;
    this->sig_index = _sig_index;
    return gpg_editkey (this->ctx, this->key, this);
}

/* Set the preferred keyserver for the given key to @url.
   If @_uid_index is -1, set the keyserver for all user-ids.
   Return value: 0 on success. */
gpgme_error_t
GpgKeyEdit::setPreferredKeyserver (int _uid_index, const char *_url)
{
    if (!this->key)
	return gpg_error (GPG_ERR_INV_OBJ);
    if (key_has_passwd && !this->pass)
	return gpg_error (GPG_ERR_INV_PASSPHRASE);
    if (!_url)
	return gpg_error (GPG_ERR_INV_ARG);

    type = GPG_EDITKEY_KEYSERV;
    this->url = _url;
    this->uid_index = _uid_index;
    return gpg_editkey (this->ctx, this->key, this);
}


/* Return the saved user-id index. */
int
GpgKeyEdit::getUseridIndex (void)
{
    return uid_index;
}


/* Return the saved key index. */   
int 
GpgKeyEdit::getKeyIndex (void)
{
    return key_index;
}


/* Return the saved sig index. */
int
GpgKeyEdit::getSigIndex (void)
{
    return sig_index;
}
