/* wptKeygenDlg.cpp - Key Generation dialog
 *	Copyright (C) 2000-2007, 2009 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <time.h>

#include "resource.h"
#include "wptTypes.h"
#include "wptNLS.h"
#include "wptGPG.h"
#include "wptCommonCtl.h"
#include "wptContext.h"
#include "wptDlgs.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptErrors.h"
#include "StringBuffer.h"


/* All valid key generation combination. */
enum gpg_keytype_t {
    GPG_KEYGEN_NONE    = 0,
    GPG_KEYGEN_DSA_ELG = 1,
    GPG_KEYGEN_DSA_RSA = 2,
    GPG_KEYGEN_RSA_RSA = 3 /*PGP*/
};


/* Generate the GPG specific genkey params with the given information.
   @keytype: See valid combinations.
   @bits: Length in bits.
   @user: user-ID name
   @comment: comment for the user-ID.
   @email: email address.
   @expdata: date of expiration or NULL.
   @passphrase: the actual passphrase.
   Return value: the gen. params. */
static char*
gpg_genkey_params (int keytype, int bits,
		   const char *user, const char *comment, const char *email,
		   const char *expdate, const char *pass)
{
    StringBuffer p;
    char *param;

    if (keytype == GPG_KEYGEN_NONE)
        return NULL;

    p = "<GnupgKeyParms format=\"internal\">\n";
    /* In this phase we set the primary key information fields. */
    switch (keytype) {
    case GPG_KEYGEN_DSA_ELG:
    case GPG_KEYGEN_DSA_RSA:
	p = p + "Key-Type: DSA\n";
	break;

    case GPG_KEYGEN_RSA_RSA:
	p = p + "Key-Type: RSA\n";
	break;

    default:
	break;
    }
    
    /* DSA v2 supports larger keys >1024 bits */
    p = p + "Key-Usage: sign\n";
    p = p + "Key-Length: " + (int)bits + "\n";

    /* The next phase is the subkey information if needed. */
    if (keytype == GPG_KEYGEN_DSA_ELG || 
	keytype == GPG_KEYGEN_DSA_RSA ||
	keytype == GPG_KEYGEN_RSA_RSA) {
	if (keytype == GPG_KEYGEN_DSA_ELG)
	    p = p + "Subkey-Type: ELG-E\n";
	else if (keytype == GPG_KEYGEN_DSA_RSA || keytype == GPG_KEYGEN_RSA_RSA)
	    p = p + "Subkey-Type: RSA\n";
	p = p + "Subkey-Usage: encrypt\n";
	p = p + "Subkey-Length: " + (int)bits + "\n";
    }

    /* Followed by the user ID information. */
    p = p + "Name-Real: " + user + "\n";
    if (comment != NULL)
	p = p + "Name-Comment: " + comment + "\n";
    p = p + "Name-Email: " + email + "\n";

    if (expdate != NULL)
	p = p + "Expire-Date: " + expdate + "\n";
    else
	p = p + "Expire-Date: 0\n";
    p = p + "Passphrase: " + pass + "\n";

    p = p + "</GnupgKeyParms>\n";
    param = p.getBufferCopy ();
    p.wipeContents ();

    return param;
}


/* Generate a key with the given params @params. @prog_cb is a user defined
   progress callback which is called during the generation.
   @fpr will store the fingerprint of the generated key.
   Return value: 0 on success. */
gpgme_error_t
gpg_genkey (const char *params, gpgme_progress_cb_t prog_cb, char **fpr)
{
    gpgme_error_t err;
    gpgme_ctx_t ctx;
    gpgme_genkey_result_t res;
    
    err = gpgme_new(&ctx);
    if (err)
        return err;
    if (prog_cb)
        gpgme_set_progress_cb (ctx, prog_cb, NULL);
    err = gpgme_op_genkey (ctx, params, NULL, NULL);
    if (!err) {
	res = gpgme_op_genkey_result (ctx);
	*fpr = res && res->fpr? m_strdup (res->fpr) : NULL;
    }
    gpgme_release (ctx);
    return err;
}



/* Reset all dialog fields to its initial value. */
static void 
reset_dlg_fields (HWND dlg)
{
    SetDlgItemText (dlg, IDC_KEYGEN_SUBKEYBITS, "");
    SetDlgItemText (dlg, IDC_KEYGEN_NAME, "");
    SetDlgItemText (dlg, IDC_KEYGEN_EMAIL, "");
    SetDlgItemText (dlg, IDC_KEYGEN_COMMENT, "");
    SetDlgItemText (dlg, IDC_KEYGEN_EXPDATE, "");
}


/* Ask the user if a keyring backup is wanted and if so,
   backup both keyrings to the selected folder. 
   Return value: true on success. */
static bool
backup_keyrings (HWND dlg)
{
    bool success = true;

    int id = msg_box (dlg, 
                 _("It is STRONGLY recommend that you backup your keyrings because they both "
                   "contain VERY important data.\nRemember that your hard disk can crash or the "
                   "files can be deleted by accident; so it is a good\nidea to store them on "
                   "a different mass stoarge like a floppy or CDR!\n\n"
                   "Backup your keyrings now?"),
                 _("WARNING - Important hint" ), MB_YESNO);
    if (id != IDYES)
	return false;

    char *path = get_gnupg_path ();
    const char *name = get_filesave_dlg (dlg, _("Destination for Public Keyring"), 
			     NULL, "pubring_bak.gpg");
    char *keyring;
    if (name != NULL) {
	keyring = make_filename (path, "pubring", "gpg");
	if (!CopyFile (keyring, name, FALSE)) {
	    log_box (_("Key Generation"), MB_ERR,
		     _("Could not copy %s -> %s"), keyring, name);
	    success = false;
	}
	free_if_alloc (keyring);    
    }

    name = get_filesave_dlg (dlg, _("Destination for Secret Keyring"), 
			     NULL, "secring_bak.gpg");
    if (name != NULL) {
	keyring = make_filename (path, "secring", "gpg");
	if (!CopyFile (keyring, name, FALSE)) {
	    log_box (_("Key Generation"), MB_ERR,
			 _("Could not copy %s -> %s"), keyring, name);
	    success = false;
	}
	free_if_alloc (keyring);   
    }
    free_if_alloc (path);
    return success;
}


/* Fill in all valid GPG algorithms. */
static void
fill_keytype_box (HWND dlg)
{
    HWND cb = GetDlgItem (dlg, IDC_KEYGEN_KEYTYPE);

#define addstr(cb, str) \
    SendMessage ((cb), CB_ADDSTRING, 0, (LPARAM)(LPCTSTR)(str))
    addstr (cb, _("DSA and ELG"));
    addstr (cb, _("DSA and RSA"));
    addstr (cb, _("RSA and RSA (PGP)") );
    SendMessage (cb, CB_SETCURSEL, 2, 0);
#undef addstr
}


time_t w32_mktime (SYSTEMTIME *st);

/* Check that the given date lies not in the past.
   Return value: 1 on success. */
int
keygen_check_date (SYSTEMTIME *st)
{	
    time_t date = w32_mktime (st);
    return date >= time (NULL)? 1 : 0;
}


/* Dialog box procedure for key generation. */
BOOL CALLBACK
keygen_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static genkey_s *ctx;
    SYSTEMTIME st;
    gpgme_error_t err;
    char *utf8_name =NULL, *utf8_comment = NULL;
    char email[128], t[64];
    char *pwd;
    char *expire = NULL, *fpr=NULL;
    int bits, use_comment, keytype = 0;
    int cancel = 0;
    char *p;
    
    switch (msg) {
    case WM_INITDIALOG:
	if (lparam != 0)
	    ctx = (genkey_s *)lparam;
        SetWindowText (dlg, _("Key Generation"));
        SetDlgItemText(dlg, IDC_KEYGEN_INFO,
			_("NOTE: Key generation can be a lengthy process! Please wait until "
			  "you get the message that key generation has finished."));
        SetDlgItemText (dlg, IDC_KEYGEN_SUBKEYINF, _("Subkey size in &bits:"));
        SetDlgItemText (dlg, IDC_KEYGEN_NAMEINF, _("&Real name:"));
        SetDlgItemText (dlg, IDC_KEYGEN_COMMINF, _("&Comment (optional):"));
        SetDlgItemText (dlg, IDC_KEYGEN_EMAILINF, _("Email &address:"));
        SetDlgItemText (dlg, IDC_KEYGEN_EXPINF, _("&Expire date:"));
        SetDlgItemText (dlg, IDC_KEYGEN_KEYTYPEINF, _("Key &type:"));
	SetDlgItemText (dlg, IDC_KEYGEN_EXPNEVER, _("&Never"));
	SetDlgItemText (dlg, IDCANCEL, _("&Cancel"));

        SetDlgItemInt (dlg, IDC_KEYGEN_SUBKEYBITS, DFAULT_KEYSIZE, FALSE);
	CheckDlgButton (dlg, IDC_KEYGEN_HIDEPWD, BST_CHECKED);
	CheckDlgButton (dlg, IDC_KEYGEN_EXPNEVER, BST_CHECKED);
	EnableWindow (GetDlgItem (dlg, IDC_KEYGEN_EXPDATE), FALSE);
        fill_keytype_box (dlg);
	center_window (dlg, NULL);
        SetForegroundWindow (dlg);
        return TRUE;

    case WM_DESTROY:
	balloon_msg_disable ();
	break;
	
    case WM_HELP:
        html_help_show(32);
	break;
	
    case WM_COMMAND:
	if (HIWORD (wparam) == CBN_SELCHANGE && 
	    LOWORD (wparam) == IDC_KEYGEN_KEYTYPE) {
	    keytype = SendMessage ((HWND)lparam, CB_GETCURSEL, 0, 0);

	    bits = DFAULT_KEYSIZE;	
	    SetDlgItemInt (dlg, IDC_KEYGEN_SUBKEYBITS, bits, FALSE);
	}

	if (HIWORD (wparam) == BN_CLICKED &&
	    LOWORD (wparam) == IDC_KEYGEN_EXPNEVER) {
	    int never = IsDlgButtonChecked (dlg, IDC_KEYGEN_EXPNEVER);
	    EnableWindow (GetDlgItem (dlg, IDC_KEYGEN_EXPDATE), !never);
	}

        switch (LOWORD (wparam)) {
        case IDOK:
	    balloon_msg_disable ();
            bits = GetDlgItemInt (dlg, IDC_KEYGEN_SUBKEYBITS, NULL, FALSE);
            if (bits < 2048 || bits > 4096) {		
		show_balloon_msg (GetDlgItem (dlg, IDC_KEYGEN_SUBKEYBITS), 
				  _("Invalid keysize value. Allowed values 2048-4096 bits."),
				  IDI_ERROR);
                return FALSE;
            }
            if (bits > DFAULT_KEYSIZE) {
                int id = msg_box (dlg, _("Do you really need such a large key?"),
				  _("Key Generation"), MB_YESNO);
                if (id == IDNO)
                    bits = DFAULT_KEYSIZE;
            }
            if (!GetDlgItemText_utf8 (dlg, IDC_KEYGEN_NAME, &utf8_name)) {
		show_balloon_msg (GetDlgItem (dlg, IDC_KEYGEN_NAME),
				  _("Please enter the name."), 
				  IDI_ERROR);
                return FALSE;
            }
	    if (strchr (utf8_name, '@')) {
		show_balloon_msg (GetDlgItem (dlg, IDC_KEYGEN_NAME),
				  _("Please do not enter the email address in the name field."),
				  IDI_WARNING);
		free_if_alloc (utf8_name);
		return FALSE;
	    }
            if (!GetDlgItemText (dlg, IDC_KEYGEN_EMAIL, email, DIM (email) -1)
                || check_email_address (email)) {
		show_balloon_msg (GetDlgItem (dlg, IDC_KEYGEN_EMAIL),
				  _("Please enter a valid email address."),
				  IDI_ERROR);
		free_if_alloc (utf8_name);
                return FALSE;
            }
            use_comment = GetDlgItemText_utf8 (dlg, IDC_KEYGEN_COMMENT, 
					       &utf8_comment);
	    if (use_comment > 0 && strchr (utf8_comment, '@')) {
		show_balloon_msg (GetDlgItem (dlg, IDC_KEYGEN_COMMENT),
				  _("Please do NOT enter the email address in the comment field."),
				  IDI_WARNING);
		free_if_alloc (utf8_name);
		free_if_alloc (utf8_comment);
		return TRUE;
	    }
            keytype = SendDlgItemMessage (dlg, IDC_KEYGEN_KEYTYPE, CB_GETCURSEL, 0, 0) + 1;
            if (IsDlgButtonChecked (dlg, IDC_KEYGEN_EXPNEVER))
                expire = NULL;
            else {
		DateTime_GetSystemtime (GetDlgItem (dlg, IDC_KEYGEN_EXPDATE), &st);
		if (!keygen_check_date (&st)) {
		    free_if_alloc (utf8_name);
		    free_if_alloc (utf8_comment);
		    show_balloon_msg (GetDlgItem (dlg, IDC_KEYGEN_EXPDATE),
				      _("The date you have chosen has already passed or is today."),
				      IDI_ERROR);
		    return TRUE;
		}
		_snprintf (t, DIM (t)-1, "%04d-%02d-%02d", st.wYear, st.wMonth, st.wDay);
                expire = t;
	    }

	    /* We don't allow empty passphrases during key generation. */
            pwd = request_passphrase2 (_("Key Generation"), 
				       PASSDLG_STRICT|PASSDLG_WARN_UTF8|PASSDLG_NOTEMPTY, &cancel);
	    if (cancel) {
		sfree_if_alloc (pwd);
		free_if_alloc (utf8_name);
		free_if_alloc (utf8_comment);
		return FALSE;
	    }
	    if (!pwd) {
                msg_box (dlg, _("Please enter the passphrase."), 
			 _("Key Generation"), MB_ERR);
		free_if_alloc (utf8_name);
		free_if_alloc (utf8_comment);
                return FALSE;
            }

	    p = gpg_genkey_params (keytype, bits, utf8_name, 
				   !use_comment && !utf8_comment? NULL :utf8_comment,
			           email, expire, pwd);
	    free_if_alloc (utf8_name);
	    free_if_alloc (utf8_comment);
            keygen_cb_dlg_create ();
            err = gpg_genkey (p, keygen_cb, &fpr);
	    sfree_if_alloc (pwd);
	    sfree_if_alloc (p); /* burn the passphrase */
            keygen_cb_dlg_destroy (1);
            if (err) {
		free_if_alloc (fpr);
                msg_box (dlg, gpgme_strerror (err), _("Key Generation"), MB_ERR);
                return FALSE;
            }
            status_box (dlg, _("Key Generation completed"), _("GnuPG Status"));
            
	    keycache_update (0, fpr);
	    keycache_update (1, fpr);
	    free_if_alloc (fpr);	    

            reset_dlg_fields (dlg);
            backup_keyrings (dlg);
	    if (ctx)
		ctx->cancel = 0;
            EndDialog (dlg, TRUE);
            return TRUE;
            
        case IDCANCEL:
	    if (ctx)
		ctx->cancel = 1;
            EndDialog (dlg, FALSE);	   
            return FALSE;
        }
        break;
    }
    
    return FALSE;
}


/* Wizard like keygen dialog for novice users. */
BOOL CALLBACK
keygen_wizard_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static genkey_s *ctx;
    static int pubkey_algo = GPG_KEYGEN_RSA_RSA;
    gpgme_error_t err;
    char email[128];
    char *utf8_name=NULL, *p, *fpr=NULL;
    char *pass = NULL;
    int cancel = 0;
    
    switch (msg) {
    case WM_INITDIALOG:
	ctx = (genkey_s *)lparam;
	ShowWindow (GetDlgItem (dlg, IDC_KEYWIZARD_USEOTHERPK), SW_HIDE);
	SetDlgItemText (dlg, IDC_KEYWIZARD_NAMEINF, _("Real name:"));
	SetDlgItemText (dlg, IDC_KEYWIZARD_EMAILINF, _("Email address:"));
	SetDlgItemText (dlg, IDC_KEYWIZARD_TITLEINF, _("Name and E-Mail Assignment"));
	SetDlgItemText (dlg, IDC_KEYWIZARD_TEXT1INF, _("Every key pair must have a name associated with it. The name and\nemail address let your correspondents know that your public key they\nare using belongs to us."));
	SetDlgItemText (dlg, IDC_KEYWIZARD_TEXT2INF, _("By associating an email address with your key pair, you will enable WinPT to assist your correspondents in selecting the correct public\nkey when communicating with you."));
	SetWindowText (dlg, _("Key Generation Wizard"));
	SetDlgItemText (dlg, IDCANCEL, _("&Cancel"));
	SetForegroundWindow (dlg);
	center_window (dlg, NULL);
	break;
	
    case WM_DESTROY:
	balloon_msg_disable ();
	break;
	
    case WM_COMMAND:
	switch (LOWORD( wparam)) {
	case IDOK:
	    balloon_msg_disable ();
	    if (!GetDlgItemText_utf8 (dlg, IDC_KEYWIZARD_NAME, &utf8_name)) {
		show_balloon_msg (GetDlgItem (dlg, IDC_KEYWIZARD_NAME),
				  _("Please enter the name."), IDI_ERROR);
		return FALSE;
	    }
	    if (strchr (utf8_name, '@')) {
		show_balloon_msg (GetDlgItem (dlg, IDC_KEYWIZARD_NAME),
				  _("Please do not enter the email address in the name field."),
				  IDI_WARNING);
		free_if_alloc (utf8_name);
		return FALSE;
	    }
	    if (!GetDlgItemText (dlg, IDC_KEYWIZARD_EMAIL, email, DIM (email)-1) ||
		check_email_address (email)) {
		show_balloon_msg (GetDlgItem (dlg, IDC_KEYWIZARD_EMAIL),
				  _("Please enter a valid email address."),
				  IDI_ERROR);
		free_if_alloc (utf8_name);
		return FALSE;
	    }
	    if (strchr (email, '<') || strchr (email, '>')) {
		show_balloon_msg (GetDlgItem (dlg, IDC_KEYWIZARD_EMAIL),
				  _("Please do not add '<' or '>' to the email address."),
				  IDI_WARNING);
		free_if_alloc (utf8_name);
		return FALSE;
	    }

	    /* We don't allow empty passphrases during key generation. */
	    pass = request_passphrase2 (_("Key Generation"),
					PASSDLG_STRICT|PASSDLG_WARN_UTF8|PASSDLG_NOTEMPTY, &cancel);
	    if (cancel) {
		free_if_alloc (utf8_name);
		return FALSE;
	    }
	    if (IsDlgButtonChecked (dlg, IDC_KEYWIZARD_USEOTHERPK))
		pubkey_algo = GPG_KEYGEN_DSA_ELG;
	    else
		pubkey_algo = GPG_KEYGEN_RSA_RSA;
	    p = gpg_genkey_params (pubkey_algo, DFAULT_KEYSIZE, utf8_name, 
				   NULL, email, NULL, pass);
	    free_if_alloc (utf8_name);
	    keygen_cb_dlg_create();
	    err = gpg_genkey (p, keygen_cb, &fpr);
	    keygen_cb_dlg_destroy (1);
	    sfree_if_alloc (p);
	    sfree_if_alloc (pass);
	    if (err) {
                msg_box (dlg, gpgme_strerror (err), 
			 _("Key Generation Wizard"), MB_ERR);
		free_if_alloc (fpr);
                return FALSE;
            }
            status_box (dlg, _("Key Generation completed"), _("GnuPG Status"));
	    keycache_update (0, fpr);
	    keycache_update (1, fpr);
	    free_if_alloc (fpr);

	    backup_keyrings (dlg);
	    if (ctx != NULL)
		ctx->cancel = 0;
	    EndDialog (dlg, TRUE);
	    break;

	case IDCANCEL:
	    if (ctx != NULL)
		ctx->cancel = 1;
	    EndDialog (dlg, FALSE);
	    break;
	}
	break;
    }
    return FALSE;
}
