/* wptMAPI.cpp - MAPI interface for sending keys.
 *	Copyright (C) 2003-2008 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <stdio.h>
#include <mapi.h>

#include "resource.h"
#include "wptTypes.h"
#include "wptErrors.h"
#include "wptW32API.h"
#include "wptGPG.h"
#include "wptVersion.h"
#include "wptCommonCtl.h"
#include "wptKeylist.h"
#include "wptKeyManager.h"


static LPMAPILOGON	    mapi_logon = NULL; 
static LPMAPILOGOFF	    mapi_logoff = NULL;     
static LPMAPISENDDOCUMENTS  mapi_send_documents = NULL; 
static LPMAPISENDMAIL	    mapi_send_mail = NULL;
static HINSTANCE	    hlib = NULL;
static int		    init = 0;

#define load_one_fnc(cast, hlib, name) (cast)GetProcAddress ((hlib), name)


/* Table with all valid MAPI errors. */
struct mapi_err_s {
    int err_no;
    const char *msg;
} mapi_errors[] = {
    {MAPI_E_FAILURE, "One or more unspecified errors occurred. "},
    {MAPI_E_AMBIGUOUS_RECIPIENT, "A recipient matched more than one of the recipient descriptor structures and MAPI_DIALOG was not set. "},
    {MAPI_E_ATTACHMENT_NOT_FOUND, "The specified attachment was not found. "},
    {MAPI_E_ATTACHMENT_OPEN_FAILURE, "The specified attachment could not be opened. "},
    {MAPI_E_BAD_RECIPTYPE, "The type of a recipient was not MAPI_TO, MAPI_CC, or MAPI_BCC. "},
    {MAPI_E_INSUFFICIENT_MEMORY, "There was insufficient memory to proceed." },
    {MAPI_E_INVALID_RECIPS, "One or more recipients were invalid or did not resolve to any address."},
    {MAPI_E_LOGIN_FAILURE, "There was no default logon, and the user failed to log on successfully when the logon dialog box was displayed."},
    {MAPI_E_TEXT_TOO_LARGE, "The text in the message was too large."},
    {MAPI_E_TOO_MANY_FILES, "There were too many file attachments."},
    {MAPI_E_TOO_MANY_RECIPIENTS, "There were too many recipients."},
    {MAPI_E_UNKNOWN_RECIPIENT, "A recipient did not appear in the address list."},
    {MAPI_E_USER_ABORT, "The user canceled one of the dialog boxes."},
    {SUCCESS_SUCCESS, "The call succeeded and the message was sent."},
    {0, NULL}
};


/* Return a human readable MAPI error based on the given error @err. */
const char*
mapi_strerror (int err)
{
    for (int i=0; mapi_errors[i].msg; i++) {
	if (err == mapi_errors[i].err_no)
	    return mapi_errors[i].msg;
    }
    return mapi_errors[0].msg;
}


/* Load MAPI library and set function pointers.
   Return value: 0 on success. */
int
mapi_init (void)
{
    if (init)
	return 0;

    hlib = LoadLibrary ("MAPI32.DLL");
    if (!hlib)
	return -1;

    mapi_logon = load_one_fnc (LPMAPILOGON, hlib, "MAPILogon");
    mapi_logoff = load_one_fnc (LPMAPILOGOFF, hlib, "MAPILogoff");
    mapi_send_documents = load_one_fnc (LPMAPISENDDOCUMENTS, hlib, 
					"MAPISendDocuments");
    mapi_send_mail = load_one_fnc (LPMAPISENDMAIL, hlib, "MAPISendMail");
    if (!mapi_logon || !mapi_logoff || !mapi_send_documents || !mapi_send_mail)
	return -1;
    init = 1;

    return 0;
}


/* Free library and cleanup. */
void
mapi_deinit (void)
{
    if (!hlib)
	return;
    FreeLibrary (hlib);
    hlib = NULL;
    init = 0;
}


/* Send the file given in @ascfile via the MAPI mechanism. */
int
mapi_send_ascfile (char *ascfile)
{
    LHANDLE hd;
    ULONG rc;

    if (!init)
	return 0;

    rc = mapi_logon (0, NULL, NULL, MAPI_LOGON_UI, 0, &hd);
    if (rc != SUCCESS_SUCCESS) {
	MessageBox (NULL, mapi_strerror (rc), _("MAPI Login failed"), MB_WARN);
	goto fail;
    }
    rc = mapi_send_documents (0, ";", ascfile, NULL, 0);    
    if (rc == MAPI_E_USER_ABORT)
	rc = SUCCESS_SUCCESS;
    if (rc != SUCCESS_SUCCESS)
	MessageBox (NULL, _("Could not send mail."), ascfile, MB_ICONERROR|MB_OK);

fail:
    mapi_logoff (hd, 0, 0, 0);
    return rc;
}


static void
free_mapi_msg (MapiMessage *msg)
{
    if (!msg)
	return;
    safe_free (msg->lpszSubject);
    safe_free (msg->lpszNoteText);
    safe_free (msg);
}


static void
free_recip_tab (MapiRecipDesc *recip, DWORD n)
{
    DWORD i;

    if (!recip || !n)
	return;
    for (i=0; i < n; i++)
	safe_free (recip[i].lpszName);
    safe_free (recip);
}


static void
free_files_tab (MapiFileDesc *files, DWORD n)
{
    DWORD i;

    if (!files)
	return;
    if (!n)
	return;
    for (i=0; i < n; i++) {
	safe_free (files[i].lpszFileName);
	safe_free (files[i].lpszPathName);
    }
    safe_free (files);
}


/* Same as mapi_send_pubkey but there is an additional note. */
int
mapi_send_pubkey_ext (winpt_key_t key, const char *keyfile, int flags)
{
    LHANDLE hd;
    MapiMessage *msg;
    MapiRecipDesc *recip;
    MapiFileDesc *attch;
    char *p, *kinf;
    const char *s;
    ULONG rc;

    if (!init)
	return 0;

    rc = mapi_logon (0, NULL, NULL, MAPI_LOGON_UI, 0, &hd);
    if (rc != SUCCESS_SUCCESS) {
	MessageBox (NULL, mapi_strerror (rc), _("MAPI Login failed"), MB_WARN);
	return rc;
    }

    msg = (MapiMessage *)calloc (1, sizeof *msg);
    msg->lpszSubject = strdup (_("OpenPGP Public Key"));

    s = _("Attached is this OpenPGP public key:\n%s\n\n"
	  "Import this key via the clipboard or the Key Manager to\n"
	  "exchange encrypted mails with the key holder and to be able\n"
	  "to verify its signatures.\n"
	  "\n"
	  "If you do not have WinPT, you can download it at http://winpt.gnupt.de");
    kinf = km_key_get_info (key, 0);
    p = (char*)malloc (strlen (s) + strlen (kinf) + 2);
    sprintf (p, s, kinf);
    free_if_alloc (kinf);

    msg->lpszNoteText = p;
    /* If the key was signed, we assume it shall be sent back to the owner. */
    if (flags) {
	recip = (MapiRecipDesc *)calloc (1, sizeof *recip);
	recip[0].ulRecipClass = MAPI_TO;
	recip[0].lpszName = strdup (key->ext->uids->uid);
	msg->lpRecips = recip;
	msg->nRecipCount = 1;
    }
    else {
	msg->lpRecips = recip = NULL;
	msg->nRecipCount = 0;
    }

    msg->nFileCount = 1;
    attch = (MapiFileDesc *)calloc (1, sizeof *attch);
    attch[0].lpszFileName = strdup (keyfile);
    attch[0].lpszPathName = strdup (keyfile);
    msg->lpFiles = attch;

    rc = mapi_send_mail (hd, 0, msg, MAPI_DIALOG , 0);
    if (rc == MAPI_E_USER_ABORT)
	rc = SUCCESS_SUCCESS;
    if (rc != SUCCESS_SUCCESS)
	MessageBox (NULL, _("Could not sent mail."), "MAPI", MB_ERR);

    free_recip_tab (recip, 1);
    free_files_tab (attch, 1);
    free_mapi_msg (msg);
    mapi_logoff (hd, 0, 0, 0);

    return 0;
}



int
mapi_send_pubkey (const char *keyid, char *keyfile)
{
    winpt_key_s key;

    memset (&key, 0, sizeof (key));
    if (winpt_get_pubkey (keyid, &key))
	return -1;
    return mapi_send_pubkey_ext (&key, keyfile, 0);
}
