/* wptNLS.cpp - W32 Native Language Support
 *	Copyright (C) 2001-2003, 2006, 2008, 2009, 2012 Timo Schulz
 *	Copyright (C) 1995-1999 Free Software Foundation, Inc.
 *
 * This code is a stripped down version of simple-gettext.c 
 * written by by Ulrich Drepper from the GPG project.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <time.h>
#include <windows.h>

#include "wptNLS.h"
#include "wptTypes.h"
#include "wptUtil.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "StringBuffer.h"

/* The magic number of the GNU message catalog format.	*/
#define MAGIC	      0x950412de
#define MAGIC_SWAPPED 0xde120495

/* Revision number of the currently used .mo (binary) file format. */
#define MO_REVISION_NUMBER 0

#define SWAPIT(flag, data) ((flag) ? do_swap_u32(data) : (data) )


/* We assume to have `unsigned long int' value with at least 32 bits. */
#define HASHWORDBITS 32

/* Header for binary .mo file format.  */
struct mo_file_header {
    DWORD magic;	    /* The magic number.	*/
    DWORD revision;	    /* The revision number of the file format.  */
    DWORD nstrings;	    /* The number of strings pairs.  */
    DWORD orig_tab_offset;  /* Offset of table with start offsets of original 
				strings.  */
    DWORD trans_tab_offset;   /* Offset of table with start offsets of translation
			     strings.  */
    DWORD hash_tab_size;	/* Size of hashing table.  */
    DWORD hash_tab_offset;    /* Offset of first hashing entry.  */
};

struct string_desc {
    DWORD length; /* Length of addressed string.  */  
    DWORD offset; /* Offset of string in file.	*/
};

struct loaded_domain_s {
    /* Full file name to the MO file */
    char *file_name;
    /* Last modification date of the MO file */
    time_t modify_time;
    
    char *data;
    int must_swap;
    DWORD nstrings;
    char *mapped;
    struct string_desc *orig_tab;
    struct string_desc *trans_tab;
    DWORD hash_size;
    DWORD *hash_tab;
};
typedef struct loaded_domain_s *gettext_domain_t;

/* List of all available languages. */
struct lang_table_s lang_list[] = {
    {"en", "English",	LANG_ENGLISH},
    {"de", "German",	LANG_GERMAN},
    {"ru", "Russian",   LANG_RUSSIAN},
    {"nl", "Dutch",     LANG_DUTCH},
    {NULL, NULL, 0}
};


/* The current gettext domain. */
static gettext_domain_t the_domain;


/* Return TRUE if the native user language is not English 
   and thus, gettext support is available */
static bool
gettext_is_required (void)
{
    LANGID lang = GetUserDefaultLangID ();
    if (PRIMARYLANGID (lang) == LANG_ENGLISH)
	return false;
    return true;
}


static DWORD
do_swap_u32 (DWORD i)
{
    return (i << 24) | ((i & 0xff00) << 8) | ((i >> 8) & 0xff00) | (i >> 24);
}


/* The so called `hashpjw' function by P.J. Weinberger
   [see Aho/Sethi/Ullman, COMPILERS: Principles, Techniques and Tools,
   1986, 1987 Bell Telephone Laboratories, Inc.]  */
static DWORD
hash_string (const char *str_param)
{
    DWORD hval, g;
    const char *str = str_param;

    hval = 0;
    while (*str != '\0') {
	hval <<= 4;
	hval += (DWORD) *str++;	
	g = hval & ((DWORD) 0xf << (HASHWORDBITS - 4));
	if (g != 0) {
	    hval ^= g >> (HASHWORDBITS - 8);
	    hval ^= g;	
	}
    }
    return hval;
}


static gettext_domain_t
load_domain (const char *filename)
{
    FILE *fp;
    size_t size;
    struct stat st;
    struct mo_file_header *data = NULL;
    gettext_domain_t domain = NULL;
    size_t to_read;
    char *read_ptr;

    fp = fopen (filename, "rb");
    if (!fp)
       return NULL;
    /* we must know about the size of the file */
    if (fstat (fileno(fp), &st) ||
	(size = (size_t)st.st_size) != (size_t)st.st_size ||
	size < sizeof (struct mo_file_header)) {
	fclose (fp);
	return NULL;
    }

    data = (struct mo_file_header *) malloc (size);
    if (!data)
	BUG (0);

    to_read = size;
    read_ptr = (char *) data;
    do {
	size_t nb = fread (read_ptr, 1, to_read, fp);
	if (nb < to_read) {
	    fclose (fp);
	    free (data);	
	    return NULL; /* read error */		
	}
	read_ptr += nb;	
	to_read -= nb;
    } while (to_read > 0);
    fclose (fp);

    /* Using the magic number we can test whether it really is a message catalog file.  */
    if (data->magic != MAGIC && data->magic != MAGIC_SWAPPED) {
	/* The magic number is wrong: not a message catalog file.  */
	free (data);
	return NULL;
    }

    domain = (struct loaded_domain_s *)calloc (1, sizeof *domain);
    if (!domain)
	BUG (0);
    domain->data = (char *) data;
    domain->must_swap = data->magic != MAGIC;

    /* Fill in the information about the available tables.  */
    switch (SWAPIT(domain->must_swap, data->revision)) {
      case 0:
	  domain->nstrings = SWAPIT (domain->must_swap, data->nstrings);
	  domain->orig_tab = (struct string_desc *)
	      ((char *) data + SWAPIT (domain->must_swap, data->orig_tab_offset));
	  domain->trans_tab = (struct string_desc *)
	      ((char *) data + SWAPIT (domain->must_swap, data->trans_tab_offset));
	  domain->hash_size = SWAPIT (domain->must_swap, data->hash_tab_size);	
	  domain->hash_tab = (DWORD *)	
	      ((char *) data + SWAPIT (domain->must_swap, data->hash_tab_offset));	
	  break;

      default: /* This is an invalid revision.	*/
	  free (data);
	  free (domain);
	  return NULL;
    }

    /* allocate an array to keep track of code page mappings */
    domain->mapped = (char *)calloc (1, domain->nstrings);
    if (!domain->mapped) 
	BUG (0);
    
    domain->file_name = strdup (filename);
    domain->modify_time = st.st_mtime;
    return domain;
}


/* To avoid problems that the current directory is not the
 * same folder the WinPT.exe is located in, we extract the
 * module directory.
 */
char*
get_module_dir (char *buf, DWORD buflen)
{
    if (buflen < 1 || !GetModuleFileName (glob_hinst, buf, buflen-1))
	return NULL;

    size_t pos = strlen (buf);
    while (pos > 0 && buf[--pos] != '\\')
	;
    buf[pos + 1] = '\0';
    return buf;
}


/* Returns a localisation name for a specific resource (file).
 * For instance, lang=DE and file=winpt.chm will become winpt.DE.chm
 */
char*
get_locale_name (const char *file)
{
    const char *lang = gettext_get_langid ();
    char buf[MAX_PATH+1];
    
    char *p = strrchr (file, '.');
    if (p == NULL) {
	p = new char[strlen(file)+strlen(lang)+2+1];
	sprintf (p, "%s.%s", file, lang);
	return p;
    }
    char *name = substr (file, 0, p - file);
    char *ext = substr (file, p - file + 1, strlen(file));
    p = get_module_dir (buf, MAX_PATH);
    if (!p)
	BUG (NULL);
    StringBuffer s = p;
    s = s + name + "." + lang + "." + ext;
    p = s.getBufferCopy();
    free_if_alloc (name);
    free_if_alloc (ext);
    return p;
}


/* Deallocate static resources. */
void
gettext_free_current_domain (void)
{
    if (the_domain == NULL)
	return;
    free (the_domain->data);
    free (the_domain->mapped);
    free (the_domain->file_name);
    free (the_domain);
    the_domain = NULL;
}


/* Replace and release the old domain and assign a new domain then */
static void
gettext_update_domain (gettext_domain_t new_dom)
{
    gettext_free_current_domain ();
    the_domain = new_dom;
}


char* get_reg_entry_mo (void);

/* Fallback code which is used when no new NLS file were found.
   Then we read the MODir registry key and check if a file named 
   'winpt.mo' exists in this directory. */
static gettext_domain_t
load_modir_domain (void)
{        
    char *modir = get_reg_entry_mo ();
    if (!modir)
	return NULL;
    if (dir_exist_check (modir)) {
	free_if_alloc (modir);
	return NULL;
    }
    StringBuffer s = modir;
    if (modir[strlen (modir)-1] != '\\')
	s += "\\";
    s += "winpt.mo";
    
    gettext_domain_t domain = NULL;
    const char *fname = s.getBuffer();
    if (!file_exist_check (fname))
	domain = load_domain (fname);
    free_if_alloc (modir);
    return domain;
}


/* Try to load the user domain, either from
   the WinPT.exe directory or from the registry */
void
gettext_set_user_domain (void)
{
    if (!gettext_is_required ())
	return;
    
    gettext_domain_t domain = NULL;
    char *file = get_locale_name ("winpt.mo");
    if (file && !file_exist_check (file))
	domain = load_domain (file);
    else
	domain = load_modir_domain ();
    free_if_alloc (file);
    gettext_update_domain (domain);
}


int
gettext_domain_needs_refresh (void)
{
    struct stat st;
    
    if (the_domain == NULL || stat (the_domain->file_name, &st))
	return 0;
    /* '1' means that the MO file were modified in the mean time
       and the application should reload the domain. */
    if (st.st_mtime != the_domain->modify_time)
	return 1;
    
    return 0;
}

static const char*
get_string (gettext_domain_t domain, DWORD idx)
{
    char *p = domain->data + SWAPIT(domain->must_swap,
				    domain->trans_tab[idx].offset);
    if (!domain->mapped[idx])
	domain->mapped[idx] = 1;
    /* FIXME: all text is in UTF-8 and need to be converted to
       the locale charset. But the caller does not free the data and
       thus we will get a leak */
    return (const char *)p;
}


const char *
gettext (const char *msgid)
{
    size_t act = 0;
    size_t top, bottom;
    
    gettext_domain_t domain = the_domain;
    if (domain == NULL)
	goto not_found;

    /* Locate the MSGID and its translation.  */
    if (domain->hash_size > 2 && domain->hash_tab) {
	/* Use the hashing table.  */		
	DWORD len = strlen (msgid);	
	DWORD hash_val = hash_string (msgid);	
	DWORD idx = hash_val % domain->hash_size;	
	DWORD incr = 1 + (hash_val % (domain->hash_size - 2));	
	DWORD nstr = SWAPIT (domain->must_swap, domain->hash_tab[idx]);

	if (!nstr) /* Hash table entry is empty.  */	
	    goto not_found;
	
	if (SWAPIT(domain->must_swap,	
	    domain->orig_tab[nstr - 1].length) == len	
	    && !strcmp (msgid,	
			domain->data + SWAPIT (domain->must_swap,
					       domain->orig_tab[nstr - 1].offset)))
	    return get_string (domain, nstr - 1);
	for(;;) {
	    if (idx >= domain->hash_size - incr)
		idx -= domain->hash_size - incr;
	    else
		idx += incr;
	    nstr = SWAPIT (domain->must_swap, domain->hash_tab[idx]);
	    if (!nstr)	
		goto not_found; /* Hash table entry is empty.  */

	    if (SWAPIT (domain->must_swap,
			domain->orig_tab[nstr - 1].length) == len
		&& !strcmp (msgid,
			    domain->data + SWAPIT (domain->must_swap,
						   domain->orig_tab[nstr - 1].offset)))
		
		return get_string (domain, nstr-1);
	} /* NOTREACHED */
    }

    /* Now we try the default method:  binary search in the sorted
       array of messages.  */
    bottom = 0;
    top = domain->nstrings;
    while (bottom < top) {	
	act = (bottom + top) / 2;	
	int cmp_val = strcmp (msgid, domain->data + SWAPIT(domain->must_swap,
							   domain->orig_tab[act].offset));
	if (cmp_val < 0)
	    top = act;
	else if (cmp_val > 0)
	    bottom = act + 1;
	else
	    return get_string (domain, act);
    }

not_found:
    return msgid;
}



/* Map the user specific language ID to a gettext conform language string.
   English "en" is defined as the fallback.
   Example: LANG_GERMAN -> "de" */
const char*
gettext_get_langid (void)
{
    LANGID lang = GetUserDefaultLangID ();
    
    for (int i=0; lang_list[i].id; i++) {
	if (PRIMARYLANGID (lang) == lang_list[i].langid)
	    return lang_list[i].id;
    }
    
    return "en";
}


/* Take a table with control item IDs and their translation
   and set each text to the translated value. */
void
gettext_localize_dialog (HWND dlg, struct gettext_tab_s *tab, const char *title)
{
    for (int i=0; tab[i].trans != NULL; i++)
	SetDlgItemText (dlg, tab[i].ctlid, tab[i].trans);
    if (title != NULL)
	SetWindowText (dlg, title);
}
