/* wptPassphraseDlg.cpp - Dialog to get the passphrase
 *	Copyright (C) 2001-2006, 2009 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>

#include "resource.h"
#include "wptGPG.h"
#include "wptCommonCtl.h"
#include "wptContext.h"
#include "wptDlgs.h"
#include "wptNLS.h"
#include "wptW32API.h"
#include "wptUTF8.h"
#include "wptVersion.h"
#include "wptTypes.h"
#include "wptRegistry.h"
#include "wptErrors.h"
#include "StringBuffer.h"


struct passphrase_s {
    char *title;    /* Title of the dialog. */
    // FIXME: do not use static buffer
    char pwd[2048];  /* The actual passphrase. */ 
    int strict;	    /* Do a simple check how good the passphrase is. */
    int repeat;	    /* Indicate last try was wrong. */   
    int cancel;	    /* 1 if user cancelled operation. */
    int warn_utf8;
    int not_empty;
    gpgme_key_t key;
};


const char* get_keyinfo (gpgme_key_t key);

static subclass_s passwd_edit_proc;

/* We sub-class the passphrase edit control to monitor
   the structure of the passphrase. This output is used
   as a quality INDICATOR(!). */
static BOOL CALLBACK
edit_subclass_proc (HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{   
    HWND dlg = passwd_edit_proc.dlg;        
    switch (msg) {
    case WM_CHAR:
	int n, pos = 0;
	n = GetWindowTextLength (GetDlgItem (dlg, IDC_PASSWD_PWD));
	if (n > 0)
	    pos = n*6;
	if (pos > 100)
	    pos = 100;
	SendDlgItemMessage (dlg, IDC_PASSWD_QOP, PBM_SETPOS, (WPARAM)pos, 0);
	break;
    }
    
    return CallWindowProc (passwd_edit_proc.old, hwnd, msg, wparam, lparam);
}


/* Fill in the key information in the combo box,
   according to the key in @key. */
static void
set_passphrase_hint (HWND dlg, gpgme_key_t key)
{
    char *uid = utf8_to_native (key->uids->name);
    
    StringBuffer p;    
    p = p + uid + " (" + get_keyinfo (key) + ")";

    SendDlgItemMessage (dlg, IDC_PASSWD_KEYINF,
			CB_ADDSTRING, 0, (LPARAM)(char *)p.getBuffer ());
    SendDlgItemMessage (dlg, IDC_PASSWD_KEYINF, CB_SETCURSEL, 0, 0);
    EnableWindow (GetDlgItem (dlg, IDC_PASSWD_KEYINF), FALSE);    
    free_if_alloc (uid);
}


/* Dialog procedure to request a passphrase from the user. */
static BOOL CALLBACK
passwd_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static passphrase_s *pwd;
    int nbytes;
    int cancel;
    
    switch (msg) {
    case WM_INITDIALOG:
        pwd = (passphrase_s *)lparam;
	if (!pwd)
	    BUG (0);
        SetWindowText (dlg, _("Passphrase Dialog"));
	CheckDlgButton (dlg, IDC_PASSWD_HIDE, BST_CHECKED);
        if (pwd->title)
            SetWindowText (dlg, pwd->title);
	if (pwd->repeat)
	    SetDlgItemText (dlg, IDC_PASSWD_INFO, _("Repeat Passphrase"));
	else
	    SetDlgItemText (dlg, IDC_PASSWD_INFO, _("Enter Passphrase"));
	if (pwd->key) {
	    ShowWindow (GetDlgItem (dlg, IDC_PASSWD_QOPINF), SW_HIDE);
	    ShowWindow (GetDlgItem (dlg, IDC_PASSWD_QOP), SW_HIDE);
	    set_passphrase_hint (dlg, pwd->key);
	}
	else
	    ShowWindow (GetDlgItem (dlg, IDC_PASSWD_KEYINF), SW_HIDE);
	SetDlgItemText (dlg, IDC_PASSWD_HIDE, _("&Hide Typing"));
	SetDlgItemText (dlg, IDCANCEL, _("&Cancel"));
	SetDlgItemText (dlg, IDOK, _("&OK"));
	SetDlgItemText (dlg, IDC_PASSWD_QOPINF, _("Quality indicator:"));
		
	if (!pwd->key) {
	    HWND hwnd;
	    hwnd = GetDlgItem (dlg, IDC_PASSWD_PWD);
	    passwd_edit_proc.dlg = dlg;
	    passwd_edit_proc.current = (WNDPROC)edit_subclass_proc;
	    passwd_edit_proc.old = (WNDPROC)GetWindowLong(hwnd, GWL_WNDPROC);
	    if (passwd_edit_proc.old) {
		if (!SetWindowLong (hwnd, GWL_WNDPROC, (LONG)passwd_edit_proc.current)) {
		    msg_box (dlg, "Could not set keylist window procedure.",
			     _("Key Manager"), MB_ERR);
		    BUG (NULL);
		}
	    }
	}

        SetFocus (GetDlgItem (dlg, IDC_PASSWD_PWD));
	center_window2 (dlg, NULL, HWND_TOPMOST);
	center_window (dlg, NULL);
        SetForegroundWindow (dlg);
        return FALSE;

    case WM_ACTIVATE:
	/* See comment in wptPassphraseCBDlg.cpp for more information. */
	if (!reg_prefs.no_safe_pwd_ctrl)
	    safe_edit_control_init (dlg, IDC_PASSWD_PWD);
	break;

    case WM_DESTROY:
	if (!reg_prefs.no_safe_pwd_ctrl)
	    safe_edit_control_free (dlg, IDC_PASSWD_PWD);
	balloon_msg_disable ();
	break;
        
    case WM_COMMAND:
	if (HIWORD (wparam) == BN_CLICKED && 
	    LOWORD (wparam) == IDC_PASSWD_HIDE) {
	    HWND hwnd = GetDlgItem (dlg, IDC_PASSWD_PWD);
	    int hide = IsDlgButtonChecked (dlg, IDC_PASSWD_HIDE);

	    SendMessage (hwnd, EM_SETPASSWORDCHAR, hide? '*' : 0, 0);
	    SetFocus (hwnd);
	}

        switch (LOWORD (wparam)) {
        case IDOK:
            pwd->cancel = 0;
	    nbytes = SafeGetDlgItemText (dlg, IDC_PASSWD_PWD, 
					 pwd->pwd, DIM (pwd->pwd)-1);
	    if (!nbytes && pwd->not_empty) {
		show_balloon_msg (GetDlgItem (dlg, IDC_PASSWD_PWD),
				  _("Please enter a passphrase."), IDI_ERROR);
		return TRUE;
	    }

            if (!nbytes)
                strcpy (pwd->pwd, "");
	    else if (pwd->strict && check_passwd_quality (pwd->pwd, 0)) {
		int id = msg_box (dlg, _("Your passphrase should be at least 8 characters"
					 " long\nand should contain non-alphabetic characters."
					 "\n\nContinue?"),
				  _("Key Generation"), MB_ICONWARNING|MB_YESNO);
		if (id == IDNO)
		    return TRUE;
	    }
	    if (pwd->warn_utf8 && !pwd->repeat &&
		     is_8bit_string (pwd->pwd)) {
		cancel = msg_box (dlg, 
			    _("The passphrase contains 8-bit characters.\n"
			      "Make sure that all systems you work on properly support UTF-8 handling.\n"
			      "Continue?"),
			_("Key Generation"), MB_ICONWARNING|MB_YESNO);
		if (cancel == IDNO) {
		    wipememory (pwd->pwd, sizeof (pwd->pwd));
		    return TRUE;
		}
	    }
            SetDlgItemText (dlg, IDC_PASSWD_PWD, "");
            EndDialog (dlg, 1);
            return TRUE;
            
        case IDCANCEL:
            pwd->cancel = 1;
            EndDialog (dlg, 0);
            return TRUE;
        }
        break;
    }
    
    return FALSE;
}


/* Same as request_passphrase(), but with an additional hint about the
   key to unprotect. */
char*
request_key_passphrase (gpgme_key_t key, const char *title, int *ret_cancel)
{
    passphrase_s pass;
    char *p;

    *ret_cancel = 0;
    memset (&pass, 0, sizeof (pass));
    pass.key = key;
    pass.title = title? m_strdup (title) : NULL;

    DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_PASSWD, glob_hwnd,
		    passwd_dlg_proc, (LPARAM)&pass);
    free_if_alloc (pass.title);
    if (pass.cancel == 1) {
	*ret_cancel = 1;
	return NULL;
    }

    p = m_strdup (pass.pwd);
    wipememory (pass.pwd, sizeof (pass.pwd));
    return p;
}


/* Request a passphrase from the user. @title is the title of the
   dialog and @ret_cancel is true if user cancelled the operation.
   Return value: the passphrase or NUL for an error. */
char*
request_passphrase (const char *title, int flags, int *ret_cancel)
{
    passphrase_s pass;
    char *p;

    *ret_cancel = 0; /* reset */
    memset (&pass, 0, sizeof (pass));
    if (title && *title)
	pass.title = m_strdup (title);
    pass.repeat = flags & PASSDLG_INIT? 0 : 1;
    pass.strict = flags & PASSDLG_STRICT? 1 : 0;
    pass.not_empty = flags & PASSDLG_NOTEMPTY? 1: 0;
    pass.warn_utf8 = flags & PASSDLG_WARN_UTF8? 1 : 0;
    DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_PASSWD, glob_hwnd,
		    passwd_dlg_proc, (LPARAM)&pass);
    free_if_alloc (pass.title);
    if (pass.cancel == 1) {
	*ret_cancel = 1;
	return NULL;
    }
    p = m_strdup (pass.pwd);
    wipememory (pass.pwd, sizeof (pass.pwd));
    return p;
}


/* Request a passphrase from the user but also confirm the passphrase
   to make sure there is no typo. Arguments same as in the normal version
   of the function. */
char*
request_passphrase2 (const char *title, int flags, int *ret_cancel)
{
    char *pass1, *pass2;
    int equal = 0, cancel = 0;

    *ret_cancel = 1;
    while (!equal) {
	pass1 = request_passphrase (title, PASSDLG_INIT|flags, &cancel);
	if (cancel)
	    return NULL;
	pass2 = request_passphrase (title, PASSDLG_REPEAT, &cancel);
	if (cancel) {
	    sfree_if_alloc (pass1);
	    return NULL;
	}

	if (strcmp (pass1, pass2)) {
	    msg_box (NULL, _("Passphrases do not match. Please try again."),
		     title, MB_INFO);
	    sfree_if_alloc (pass1);
	    sfree_if_alloc (pass2);
	}
	else
	    equal = 1;
    }
    sfree_if_alloc (pass2);
    *ret_cancel = 0;
    return pass1;
}
