/* wptPreferencesDlg.cpp - Dialog for the preferences
 *	Copyright (C) 2001-2003, 2005-2006, 2008-2009 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <shlobj.h>

#include "wptNLS.h"
#include "wptGPG.h"
#include "wptCommonCtl.h"
#include "wptDlgs.h"
#include "wptTypes.h"
#include "wptErrors.h"
#include "wptRegistry.h"
#include "wptHotkey.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptAgent.h"
#include "wptKeyserver.h"
#include "resource.h"


/* Dialog IDs of all hotkeys. */
static int hotkeys[] = {
    /* clipboard based */
    // FIXME: why the wrong order leads to a correct storing
    // in the registry!? (2,1) should be (1,2)
    IDC_PREFS_CLIP_DECRYPT_VERIFY,
    IDC_PREFS_CLIP_ENCRYPT,    
    IDC_PREFS_CLIP_SIGN,
    IDC_PREFS_CLIP_SIGNENC,
    /* current window based */
    IDC_PREFS_CWS_ENCRYPT,
    IDC_PREFS_CWS_DECRYPT_VERIFY,
    IDC_PREFS_CWS_SIGN,
    IDC_PREFS_CWS_SIGNENC,
    0
};


/* Check that the given hotkey is in A..Z or a..z. */
static int
check_hotkey (char *key)
{
    if (*key >= 'A' && *key <= 'Z')
        return 1;
    else if (*key >= 'a' && *key <= 'z') {
        *key = *key - 32;
        return 1;
    }
    return 0;
}


/* Disable all hotkey controls in the dialog @dlg if val is 1. */
static void
disable_hotkey_items (HWND dlg, int val)
{
    int mode = val? FALSE : TRUE;
    
    for (int i=0; hotkeys[i] != 0; i++)
	EnableWindow (GetDlgItem (dlg, hotkeys[i]), mode);
}



/* En- or disable the backup items in the dialog @dlg.
   If val != 0 activate them, disable them otherwise. */
static void
enable_backup_items (HWND dlg, int val)
{
    int mode = val? TRUE : FALSE;

    EnableWindow (GetDlgItem (dlg, IDC_PREFS_BAKHOME), mode);
    EnableWindow (GetDlgItem (dlg, IDC_PREFS_BAKUSER), mode);
    EnableWindow (GetDlgItem (dlg, IDC_PREFS_BAKSELECT), mode);
    EnableWindow (GetDlgItem (dlg, IDC_PREFS_BAKPATH), mode);
    EnableWindow (GetDlgItem (dlg, IDC_PREFS_BAKSECRING), mode);
}

/* Helper function to prepare dialog. */
static void
on_init_dialog (HWND dlg)
{
    struct gettext_tab_s prefs[] = {
	{IDC_PREFS_VIEWER, _("Use clipboard &viewer to display the plaintext")},
	{IDC_PREFS_WWINFO, _("Word wrap cleartext &signatures at column:")},
	{IDC_PREFS_DISABLE_HOTKEYS, _("&Disable hotkeys (Not recommended!)")},
	{IDC_PREFS_TRUST, _("Skip key validation and assume that keys are always fully trusted")},
	{IDC_PREFS_AUTOBACKUP, _("&Automatic keyring backup when WinPT closes")},
	{IDC_PREFS_BAKHOME, _("Backup to &keyring folder")},
	{IDC_PREFS_BAKUSER, _("Backup to:")},
	{IDC_PREFS_KEYLISTINF, _("Select &key list mode")},
	{IDC_PREFS_KSINF, _("Keyserver &config")},
	{IDC_PREFS_CACHEINF, _("Cache &passphrases for 'n' minutes:")},
	{IDC_PREFS_CACHEHINT, _("(CTRL+ALT+F to clear the cache)")},
	{IDC_PREFS_ALLOPTINF, _("General options")},
	{IDC_PREFS_CLIPINF, _("Clipboard hotkeys CTRL+ALT")},
	{IDC_PREFS_CURRINF, _("Current window hotkeys CTRL+SHIFT")},
	{IDC_PREFS_KSELFILE, _("Browse...")},
	{IDC_PREFS_BAKSELECT, _("Browse...")},
	{IDC_PREFS_EXTINF, _("Default extension for encrypted files:")},
	{IDC_PREFS_BAKSECRING, _("&Backup includes secret keyring")},
	
	{IDC_PREFS_CLIP_HT_ENC, _("Encrypt")},
	{IDC_PREFS_CLIP_HT_DEC, _("Decrypt/Verify")},
	{IDC_PREFS_CLIP_HT_SIG, _("Sign")},
	{IDC_PREFS_CLIP_HT_SAE, _("Sign && Encrypt")},	
	{IDC_PREFS_CW_HT_ENC, _("Encrypt")},
	{IDC_PREFS_CW_HT_DEC, _("Decrypt/Verify")},
	{IDC_PREFS_CW_HT_SIG, _("Sign")},
	{IDC_PREFS_CW_HT_SAE, _("Sign && Encrypt")},
	{IDCANCEL, _("&Cancel")},
	{0, NULL}};
    int i;
    HWND cb;

    gettext_localize_dialog (dlg, prefs, _("WinPT Preferences"));

    SetDlgItemInt (dlg, IDC_PREFS_CACHETIME, reg_prefs.cache_time/60, TRUE);
    SetDlgItemInt (dlg, IDC_PREFS_WORDWRAP, reg_prefs.word_wrap, TRUE);
    if (reg_prefs.backup.path)
	SetDlgItemText (dlg, IDC_PREFS_BAKPATH, reg_prefs.backup.path);
	
    for (i=0; hotkeys[i]; i++) {
	if (!wpt_hotkeys[i].enabled)
	    continue;
	char key[2];
	key[0] = wpt_hotkeys[i].key;
	key[1] = 0;
	SetDlgItemText (dlg, hotkeys[i], key);
    }
    cb = GetDlgItem (dlg, IDC_PREFS_DEFEXT);
    combox_add_string (cb, "GnuPG (.gpg)");
    combox_add_string (cb, "PGP   (.pgp)");
    SendMessage (cb, CB_SETCURSEL, 0, 0);

    CheckDlgButton (dlg, IDC_PREFS_DISABLE_HOTKEYS,
		    reg_prefs.no_hotkeys ? BST_CHECKED : BST_UNCHECKED);
    CheckDlgButton (dlg, IDC_PREFS_VIEWER,
		    reg_prefs.use_viewer ? BST_CHECKED: BST_UNCHECKED);
    CheckDlgButton (dlg, IDC_PREFS_TRUST,
		    reg_prefs.always_trust? BST_CHECKED : BST_UNCHECKED);
    CheckDlgButton (dlg, IDC_PREFS_AUTOBACKUP,
		    reg_prefs.auto_backup? BST_CHECKED : BST_UNCHECKED);
    CheckDlgButton (dlg, IDC_PREFS_BAKHOME, 
		    reg_prefs.backup.mode==1? BST_CHECKED : BST_UNCHECKED);
    CheckDlgButton (dlg, IDC_PREFS_BAKUSER,
		    reg_prefs.backup.mode==2? BST_CHECKED : BST_UNCHECKED);
    CheckDlgButton (dlg, IDC_PREFS_BAKSECRING, 
		    reg_prefs.backup.include_secr? BST_CHECKED :BST_UNCHECKED);
    SendDlgItemMessage (dlg, IDC_PREFS_DEFEXT, CB_SETCURSEL, 
			reg_prefs.default_ext, 0);
    if (reg_prefs.no_hotkeys)
        disable_hotkey_items (dlg, 1);
    if (!reg_prefs.auto_backup)
	enable_backup_items (dlg, 0);
    EnableWindow (GetDlgItem (dlg, IDC_PREFS_BAKPATH), 
		      reg_prefs.backup.mode==1?FALSE : TRUE);
    EnableWindow (GetDlgItem (dlg, IDC_PREFS_BAKSELECT),
		      reg_prefs.backup.mode==1? FALSE : TRUE);
    center_window (dlg, NULL);    
    SetForegroundWindow (dlg);
}


/* Dialog box procedure for the WinPT preferences. */
BOOL CALLBACK
prefs_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    int rc;
    char t[3];
    char path[256], *p;
    
    switch (msg) {
    case WM_INITDIALOG:
	on_init_dialog (dlg);
        return TRUE;
        
    case WM_DESTROY:
	balloon_msg_disable ();
	break;
	
    case WM_COMMAND:
        switch (HIWORD (wparam)) {
        case BN_CLICKED:
	    switch ((int)LOWORD (wparam)) {
	    case IDC_PREFS_DISABLE_HOTKEYS:            
                reg_prefs.no_hotkeys ^= 1;
                disable_hotkey_items (dlg, reg_prefs.no_hotkeys);
		break;

	    case IDC_PREFS_TRUST:
		reg_prefs.always_trust ^= 1;
		if (reg_prefs.always_trust) {
		    msg_box (dlg,
		_("In most cases it is not a good idea to enable this setting.\n"
		  "If you know what you are doing, you can enable this flag, otherwise\n"
		  "it is safe to leave this flag untouched."), 
		  _("Preferences"), MB_WARN);
		}
		break;

	    case IDC_PREFS_AUTOBACKUP:
		reg_prefs.auto_backup ^= 1;		    
		enable_backup_items (dlg, reg_prefs.auto_backup);
		if (reg_prefs.auto_backup == 1) {
		    if (!IsDlgButtonChecked (dlg, IDC_PREFS_BAKHOME) &&
			!IsDlgButtonChecked (dlg, IDC_PREFS_BAKUSER))
			CheckDlgButton (dlg, IDC_PREFS_BAKHOME, BST_CHECKED);
		}
		if (IsDlgButtonChecked (dlg, IDC_PREFS_BAKHOME))
		    EnableWindow (GetDlgItem (dlg, IDC_PREFS_BAKPATH), FALSE);
		break;

	    case IDC_PREFS_BAKHOME:		
		EnableWindow (GetDlgItem (dlg, IDC_PREFS_BAKSELECT), FALSE);
		EnableWindow (GetDlgItem (dlg, IDC_PREFS_BAKPATH), FALSE);
		break;

	    case IDC_PREFS_BAKUSER:
		EnableWindow (GetDlgItem (dlg, IDC_PREFS_BAKSELECT), TRUE);
		EnableWindow (GetDlgItem (dlg, IDC_PREFS_BAKPATH), TRUE);
		break;
	    }
            break;
        }
        switch (LOWORD (wparam)) {
	case IDC_PREFS_BAKSELECT:
	    const char *bpath;
	    bpath = get_folder_dlg (dlg, _("Select GPG backup path"), NULL);
	    if (bpath)
		SetDlgItemText (dlg, IDC_PREFS_BAKPATH, bpath);
	    break;

        case IDOK:
            rc = GetDlgItemInt (dlg, IDC_PREFS_CACHETIME, NULL, FALSE);
            if (rc <= 0) {
                reg_prefs.cache_time = 0;
		agent_flush_cache ();
	    }
            else if (rc > 720) {
		show_balloon_msg (GetDlgItem (dlg, IDC_PREFS_CACHETIME),
				  _("Please enter a value that is between 1-720.\nIt is not "
				    "a good idea to cache the passphrase more than 12 hours."),
				  IDI_ERROR);
		SetDlgItemInt (dlg, IDC_PREFS_CACHETIME, 0, FALSE);
		return FALSE;
            }
	    if (reg_prefs.cache_time != rc)
		agent_flush_cache ();
            reg_prefs.cache_time = rc*60;
            rc = GetDlgItemInt (dlg, IDC_PREFS_WORDWRAP, NULL, FALSE);
	    if (rc <= 0)
                reg_prefs.word_wrap = 0;
            else if (rc > 80) {
		show_balloon_msg (GetDlgItem (dlg, IDC_PREFS_WORDWRAP),
				  _("Please enter a value between 1-80."),
				  IDI_ERROR);
                return FALSE;
            }
	    reg_prefs.default_ext = SendDlgItemMessage (dlg, IDC_PREFS_DEFEXT, CB_GETCURSEL, 0, 0);
            reg_prefs.word_wrap = rc;
	    reg_prefs.use_viewer = IsDlgButtonChecked (dlg, IDC_PREFS_VIEWER)? 1 : 0;
	    reg_prefs.backup.mode = IsDlgButtonChecked (dlg, IDC_PREFS_BAKHOME ) ? 1 :
				    IsDlgButtonChecked (dlg, IDC_PREFS_BAKUSER)? 2 : 0;
	    reg_prefs.backup.include_secr = IsDlgButtonChecked (dlg, IDC_PREFS_BAKSECRING)? 1 : 0;
	    if (reg_prefs.backup.mode == 2) {
		if (!GetDlgItemText (dlg, IDC_PREFS_BAKPATH, path, DIM (path)-1)
		    || dir_exist_check (path)) {
		    show_balloon_msg (GetDlgItem (dlg, IDC_PREFS_BAKPATH),
				      _("The specified backup folder is invalid."),
				      IDI_ERROR);
		    return FALSE;
		}
		free_if_alloc (reg_prefs.backup.path);
		p = reg_prefs.backup.path = m_strdup (path);
	    }
            
            if (IsDlgButtonChecked (dlg, IDC_PREFS_DISABLE_HOTKEYS)) {
                hotkeys_unregister (glob_hwnd);
                reg_prefs.no_hotkeys = 1;
            }
            else {
                reg_prefs.no_hotkeys = 0;
		for (int i = 0; hotkeys[i] != 0; i++) {
		    rc = GetDlgItemText (dlg, hotkeys[i], t, DIM (t)-1);
		    if (rc > 0 && check_hotkey (&t[0]))
			hotkey_enable (&wpt_hotkeys[i], t);
		    else
			hotkey_disable (&wpt_hotkeys[i]);
		}
		set_reg_entry (HKEY_CURRENT_USER, "Software\\WinPT", "DisableHotkeys", "0");
		reg_prefs.no_hotkeys = 0;
	    }

            rc = set_reg_winpt_prefs (&reg_prefs);
	    if (rc)
                msg_box (dlg, winpt_strerror (rc), _("Preferences"), MB_ERR);
            
	    if (reg_prefs.no_hotkeys == 0) {
		hotkeys_unregister (glob_hwnd);
		rc = hotkeys_register (glob_hwnd);
		if (rc)
		    msg_box (NULL, winpt_strerror (rc), _("Hotkeys"), MB_ERR);
	    }
            EndDialog (dlg, TRUE);
            return TRUE;
            
        case IDCANCEL:
	    /* Reset backup mode if no mode was chosen. */
	    if (reg_prefs.auto_backup == 1 &&
		!IsDlgButtonChecked (dlg, IDC_PREFS_BAKHOME) &&
		!IsDlgButtonChecked (dlg, IDC_PREFS_BAKUSER))
		reg_prefs.auto_backup = 0;
            EndDialog (dlg, FALSE);
            return FALSE;
        }
        break;
    }
    
    return FALSE;
}
