/* wptRegistry.cpp - Windows Registry access
 *	Copyright (C) 2000-2006 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <stdio.h>

#include "wptErrors.h"
#include "wptW32API.h"
#include "wptGPG.h"
#include "wptRegistry.h"
#include "wptKeyserver.h"
#include "wptTypes.h"
#include "wptNLS.h"
#include "wptVersion.h"
#include "wptHotkey.h"

#define rc_ok(rc) ((rc) == ERROR_SUCCESS)

/* WinPT registry path. */
#define WINPT_REG "Software\\WinPT"

/* GPG registry path. */
#define GPG_REG "Software\\GNU\\GnuPG"

#define GPG_MO_REG "Control Panel\\MingW32\\NLS"

/* GPG file association context. */
struct gpg_filetype_s {    
    const char *descr;	/* description. */
    const char *ext;	/* extension. */
    int nicon;		/* number of the image icon. */
};
typedef struct gpg_filetype_s *gpg_filetype_t;

/* Global WinPT registry prefereneces. */
struct winpt_prefs_s reg_prefs;

/* All valid configuration commands. */
const char *cfg [] = {
    NULL,	
    "CacheTime",
    "WordWrap",
    "DefaultExt",
    "Viewer",
    "AlwaysTrust",
    "AutoBackup",
    "BackupMode",
    "DisableHotkeys",	
    "NoCompressMultiMedia",	
    "Expert",
    "FMProgressBar",
    "BackupSecring"
};

/* Return != 0 if GPG4win is installed. */
int
is_gpg4win_installed (void)
{
    char *p;

    p = get_reg_entry_gpg4win (NULL);
    if (!p)
	return 0;
    if (dir_exist_check (p)) {
	free_if_alloc (p);
	return 0;
    }
    free_if_alloc (p);
    return -1;
}


/* Return != 0 if GPGee is installed. */
static int
is_gpgee_installed (void)
{
    HKEY hk;
    LONG ec;

    ec = RegOpenKeyEx (HKEY_CURRENT_USER, "Software\\GPGee", 0, KEY_READ, &hk);
    if (ec == ERROR_SUCCESS) {
	RegCloseKey (hk);
	return -1;
    }

    return 0;
}


/* Free all members in the registry preference struct. */
void
free_reg_prefs (void)
{
    free_if_alloc (reg_prefs.backup.path);
    free_if_alloc (reg_prefs.kserv_conf);
    free_if_alloc (reg_prefs.homedir);
    memset (&reg_prefs, 0, sizeof (reg_prefs));
}



/* Install the GPG related into the W32 resgistry, if the entry already
   exists the function returns immediately. */
int
regist_inst_gnupg (int create_mokey)
{
    int rc;
    HKEY reg;

    rc = RegOpenKeyEx (HKEY_CURRENT_USER, GPG_REG, 0, KEY_READ, &reg);
    if (rc == ERROR_SUCCESS) {
	RegCloseKey (reg);
	return 0;	
    }
    rc = RegCreateKey (HKEY_CURRENT_USER, GPG_REG, &reg);
    if (rc != ERROR_SUCCESS)
	return WPTERR_REGISTRY;	
    RegCloseKey (reg);

    if (!create_mokey)
	return 0;

    rc = RegOpenKeyEx (HKEY_CURRENT_USER,  GPG_MO_REG, 0, KEY_READ, &reg);
    if (rc == ERROR_SUCCESS) {
	RegCloseKey (reg);
	return 0;	
    }
    rc = RegCreateKey (HKEY_CURRENT_USER, GPG_MO_REG, &reg);
    if (rc != ERROR_SUCCESS)
	return WPTERR_REGISTRY;

    RegCloseKey (reg);
    return 0;
}


/* Create a new filetype in the W32 registry.
   We should really care of errors! Otherwise we can damage the registry! */
static int
create_file_type (const char *exefile, const char *ext, 
		  const char *extname, char *iconfile)
{
    HKEY reg = NULL;
    char deficon[MAX_PATH+1];
    char defexec[MAX_PATH+1];
    char p_exefile[MAX_PATH+1];
    int rc;
    
    rc = RegCreateKey (HKEY_CLASSES_ROOT, ext, &reg);
    if( rc_ok( rc ) ) 
	rc = RegSetValueEx( reg, NULL, 0, REG_SZ, (BYTE *)extname, strlen( extname ) );
    if( rc_ok( rc ) )
	rc = RegCloseKey( reg );
    if( rc_ok( rc ) )    
	rc = RegCreateKey( HKEY_CLASSES_ROOT, extname, &reg );
    if( rc_ok( rc ) )
	rc = RegSetValueEx( reg, NULL, 0, REG_SZ, (BYTE *) extname, strlen( extname ) );
    if( rc_ok( rc ) )
	rc = RegCloseKey( reg );
    if( !rc_ok( rc ) ) {
	rc = WPTERR_REGISTRY;
	goto leave;
    }
    
    memset( &deficon, 0, sizeof deficon );
    _snprintf( deficon, DIM (deficon) - 1, "%s\\DefaultIcon", extname );
    memset( &defexec, 0, sizeof defexec );
    _snprintf( defexec, DIM (defexec) - 1, "%s\\shell\\open\\command", extname );	
    memset( &p_exefile, 0, sizeof p_exefile );
    _snprintf( p_exefile, DIM (p_exefile) - 1, "%s %%1", exefile );
    
    rc = RegCreateKey( HKEY_CLASSES_ROOT, deficon, &reg );
    if( rc_ok( rc ) )
	rc = RegSetValueEx(reg, NULL, 0, REG_SZ, (BYTE *)iconfile, strlen( iconfile ) );
    if( rc_ok( rc ) )
	rc = RegCloseKey( reg );
    if( rc_ok( rc ) )
	rc = RegCreateKey( HKEY_CLASSES_ROOT, defexec, &reg );
    if( rc_ok( rc ) )
	rc = RegSetValueEx( reg, NULL, 0, REG_SZ, (BYTE *)p_exefile, strlen( exefile ) );
    if( rc_ok( rc ) )
	rc = RegCloseKey( reg );
    if( !rc_ok( rc ) ) {
	rc = WPTERR_REGISTRY;
	goto leave;	
    }
    
leave:
    if (reg)
	RegCloseKey (reg);
    return rc;
}


/* Register the given WinPT filetype. */
static int
regist_single_filetype (gpg_filetype_t gfile)
{
    char icon[MAX_PATH+1] = {0};
    char prog[MAX_PATH+1] = {0};

    if (!GetModuleFileName (glob_hinst, prog, DIM (prog)-1))
	return WPTERR_REGISTRY;
    _snprintf (icon, DIM (icon) -1, "%s,%d", prog, gfile->nicon);
    return create_file_type (prog, gfile->ext, gfile->descr, icon);
}

/* Install WinPT into the W32 registry, if the entry already 
   exists the function returns immediately. @with_ext can be
   used to register some file types (if 1). @created contains
   1 if the registry key was created.
   Return value: 0 on success. */
int
regist_inst_winpt (int with_ext, int *created)
{
    struct gpg_filetype_s gpg_filetypes[] = {
	{_("GPG Detached Signature"), ".sig", 1},
	{_("GPG Encrypted Data"),     ".gpg", 2},
	{_("GPG Armored Data"),       ".asc", 2},
	{0}
    };
    HKEY reg;
    char *p = NULL;
    char modpath[MAX_PATH+1];
    int rc, i, id, n;

    if (created)
	*created = 0;

    /* If GPGee is already installed deactivate file extension registering
       because it is very likely this would overwrite the GPGee settings. */
    if (is_gpgee_installed ())
	with_ext = 0;

    rc = RegOpenKeyEx (HKEY_CURRENT_USER, WINPT_REG, 0, KEY_READ, &reg);
    if (rc == ERROR_SUCCESS) {
	RegCloseKey (reg);
	rc = RegOpenKeyEx (HKEY_CURRENT_USER, WINPT_REG"\\Keyserver", 0, KEY_READ, &reg);
	if (rc != ERROR_SUCCESS) {
	    rc = RegCreateKey (HKEY_CURRENT_USER, WINPT_REG"\\Keyserver", &reg);
	    if (rc == ERROR_SUCCESS)
		RegCloseKey (reg);
	}
	else
	    RegCloseKey (reg);
	p = get_reg_entry_keyserver ("Default");
	if (!p) {
	    char buf[16];

	    _snprintf (buf, DIM (buf)-1, "%d", HKP_PORT);
	    set_reg_entry_keyserver ("Default_Port", buf);
	    set_reg_entry_keyserver ("Default", DEF_HKP_KEYSERVER);
	}
	free_if_alloc (p);
	return 0;
    }
    rc = RegCreateKey (HKEY_CURRENT_USER, WINPT_REG, &reg);
    if (rc != ERROR_SUCCESS)
	return WPTERR_REGISTRY;
    if (created)
	*created = 1;
    RegCloseKey (reg);
    if (with_ext) {
	id = msg_box (NULL, _("WinPT can register some GPG file types for you so they can "
	                      "be processed with a double click in the explorer.\n"
			      "Do you want to continue?"), _("WinPT"), MB_YESNO|MB_INFO);
	if (id == IDYES) {
	    for (i = 0; gpg_filetypes[i].ext; i++) {
		rc = RegOpenKeyEx (HKEY_CLASSES_ROOT, gpg_filetypes[i].ext, 0, KEY_READ, &reg);
		if (rc == ERROR_SUCCESS) {
		    RegCloseKey (reg);
		    id = log_box (_("WinPT WARNING"), MB_YESNO|MB_INFO,
				  _("It seems there was already a '%s' file type registered by another application.\n"
				    "Do you want to overwrite it?"), gpg_filetypes[i].ext);
		    if (id == IDNO)
			continue;
		}
		regist_single_filetype (&gpg_filetypes[i]);
	    }
	}
    }
    /* Store the install directory for later use. */
    if ((n=GetModuleFileName (NULL, modpath, MAX_PATH-1)) > 0) {
	while (n-- > 0) {
	    if (modpath[n] == '\\') {
		modpath[n] = 0;
		break;
	    }
	}
	set_reg_entry (HKEY_CURRENT_USER, WINPT_REG, "Install Directory", modpath);
    }
    return 0;
}


/* Expand a string with %foo% entries so that %foo% will
   be replaced with its actual value. */
static char*
expand_path (const char *path)
{
    DWORD len;
    char *p;

    len = ExpandEnvironmentStrings (path, NULL, 0);
    if (!len)
	return NULL;
    len += 1;
    p = new char[len+1];
    if (!p)
	return NULL;
    len = ExpandEnvironmentStrings (path, p, len);
    if (!len) {
	free_if_alloc (p);
	return NULL;
    }
    return p; 
}


/* Retrieve a registry entry with the directory given in @dir
   and the key given in @key.
   Return value is the value or NULL otherwise. */
char*
get_reg_entry (HKEY root_key, const char *dir, const char *key)
{
    HKEY reg_key = NULL;
    char *p = NULL, *pp;    
    DWORD type = REG_SZ, nbytes = 0;
    int rc;

    rc = RegOpenKeyEx (root_key, dir, 0, KEY_QUERY_VALUE, &reg_key);
    if (rc != ERROR_SUCCESS)
	goto leave;
    rc = RegQueryValueEx (reg_key, key, NULL, &type, NULL, &nbytes);
    if (rc != ERROR_SUCCESS)
	goto leave;
    if (!nbytes)
	goto leave; /* empty */
    p = new char[nbytes+1];
    if (!p)
	BUG (0);
    rc = RegQueryValueEx (reg_key, key, NULL, &type, (BYTE*)p,  &nbytes);
    if (rc != ERROR_SUCCESS)
	goto leave;
    if (type == REG_EXPAND_SZ && strchr (p, '%')) {
	pp = p;
	p = expand_path (pp);
	free_if_alloc (pp);
    }

leave:
    if (reg_key != NULL)
	RegCloseKey (reg_key);
    return p;
}


/* XXX: use REG_EXPAND_SZ to support multi user environments.
        keep this type and do not overwrite it with REG_SZ. */ 

/* Set a registry entry. */
int
set_reg_entry (HKEY root_key, const char *dir, const char *key,
	       const char *value)
{
    HKEY reg_key;
    int rc;

    rc = RegOpenKeyEx (root_key, dir, 0, KEY_WRITE, &reg_key);
    if (rc != ERROR_SUCCESS)
	return WPTERR_REGISTRY;
    rc = RegSetValueEx (reg_key, key, 0, REG_SZ, (BYTE *)value, strlen (value));
    if (rc != ERROR_SUCCESS)
	rc = WPTERR_REGISTRY;	
    RegCloseKey (reg_key);
    return rc;
}


int
set_reg_key (HKEY root_key, const char *dir, const char *key,
	     const char *value)
{    
    HKEY reg_key;
    int rc;
    
    rc = RegOpenKeyEx (root_key, dir, 0, KEY_WRITE, &reg_key);
    if (rc != ERROR_SUCCESS)
	return WPTERR_REGISTRY;
    
    rc = RegSetValueEx (reg_key, key, 0, REG_SZ, (BYTE *)value, strlen (value));
    if (rc != ERROR_SUCCESS) {
	if (RegCreateKey (root_key, key, &reg_key) != ERROR_SUCCESS) {
	    rc = WPTERR_REGISTRY;
	    goto leave;
	}
	rc = RegSetValueEx (reg_key, key, 0, REG_SZ, (BYTE *)value, strlen (value));
	if (rc != ERROR_SUCCESS)
	    rc = WPTERR_REGISTRY;
    }
    
leave:
    RegCloseKey (reg_key);
    return rc;
}


int
set_reg_entry_gpg (const char *key, const char *value)
{	
    return set_reg_entry (HKEY_CURRENT_USER, GPG_REG, key, value);
}



char*
get_reg_entry_gpg (const char *key)
{
    char *p;

    p = get_reg_entry (HKEY_CURRENT_USER, GPG_REG, key);
    if (!p || strlen (p) == 0) {
	free_if_alloc (p);
	return NULL;
    }
    return p;
}


/* Return if possible the GPG4Win installation directory concatenated
   with the string in @path if given. */
char*
get_reg_entry_gpg4win (const char *path)
{
    const char *fmt;
    char *p, *pp;

    p = get_reg_entry (HKEY_LOCAL_MACHINE, GPG_REG, "Install Directory");
    if (!p)
	return NULL;
    if (!path)
	return p;
    fmt = "%s\\%s";
    pp = new char[strlen (p) + strlen (path) + strlen (fmt) + 1];
    if (!pp)
	BUG (NULL);
    sprintf (pp, fmt, p, path);
    free_if_alloc (p);
    return pp;
}

    
char*
get_reg_entry_mo (void)
{	
    char *p, *pp;
    const char *fmt;
    const char *lang;

    p = get_reg_entry (HKEY_CURRENT_USER, GPG_MO_REG, "MODir");
    if (p)
	return p;

    lang = gettext_get_langid ();
    if (!lang)
	return NULL;
    fmt = "share\\locale\\%s\\LC_MESSAGES";
    pp = new char[strlen (fmt) + strlen (lang) + 4 + 1];
    if (!pp)    
	BUG (NULL);
    sprintf (pp, fmt, lang);
    p = get_reg_entry_gpg4win (pp);
    free_if_alloc (pp);
    return p;
}


int
set_reg_winpt_single (int id, int val)
{
    char buf[64];
    int rc;

    _snprintf (buf, DIM (buf)-1, "%d", val);
    rc = set_reg_entry (HKEY_CURRENT_USER, WINPT_REG, cfg[id], buf);
    return rc;
}


int
get_reg_winpt_single (int id)
{
    char *buf;
    int val = 0;

    buf = get_reg_entry (HKEY_CURRENT_USER, WINPT_REG, cfg[id]);
    if (buf && *buf != ' ')
	val = 1;
    else if (!buf)
	val = -1;
    free_if_alloc (buf);
    return val;
}


/* Saves the winpt preferences in the registry. */
int
set_reg_winpt_prefs (winpt_prefs_t opt)
{
    char buf[128];
    size_t i;
    int rc = 0;

    for (i=1; i < DIM (cfg); i++) {
	switch (i) {
	case CFG_CACHETIME:
	    sprintf (buf, "%d", opt->cache_time);
	    break;
	    
	case CFG_WORDWRAP:
	    sprintf (buf, "%d", opt->word_wrap);
	    break;
	    
	case CFG_FILEEXT:
	    sprintf (buf, "%d", opt->default_ext);
	    break;
	    
	case CFG_NOZIP_MMEDIA:
	    sprintf (buf, "%d", opt->no_zip_mmedia);
	    break;
	    
	case CFG_VIEWER:
	    sprintf (buf, "%d", opt->use_viewer);
	    break;
	    
	case CFG_ALWAYSTRUST:
	    sprintf (buf, "%d", opt->always_trust);
	    break;

	case CFG_AUTOBACKUP:
	    sprintf (buf, "%d", opt->auto_backup);
	    break;
	    
	case CFG_AUTOBAKMODE:
	    sprintf (buf, "%d", opt->backup.mode);
	    break;
	    
	case CFG_DISHOTKEYS:
	    sprintf (buf, "%d", opt->no_hotkeys);
	    break;
	    
	case CFG_EXPERT:
	    sprintf (buf, "%d", opt->expert);
	    break;

	case CFG_FM_PROGRESS:
	    sprintf (buf, "%d", opt->fm.progress);
	    break;

	case CFG_BACKUP_INC_SKR:
	    sprintf (buf, "%d", opt->backup.include_secr);
	    break;
	}
	rc = set_reg_entry (HKEY_CURRENT_USER, WINPT_REG, cfg[i], buf);
	if (rc)
	    goto leave;
    }

    if (opt->backup.path) {
	rc = set_reg_entry (HKEY_CURRENT_USER, WINPT_REG, "BackupPath",
			    opt->backup.path);
	if (rc)
	    goto leave;
    }

    for (i=0; wpt_hotkeys[i].name != NULL; i++) {
	if (wpt_hotkeys[i].enabled == 1) {
	    buf[0] = wpt_hotkeys[i].key;
	    buf[1] = 0;
	}
	else
	    strcpy (buf, " ");
	rc = set_reg_key (HKEY_CURRENT_USER, WINPT_REG, 
			  wpt_hotkeys[i].name, buf);
	if (rc)
	    break;
    }
    
leave:
    if (rc) {
	msg_box (NULL, _("Could not write to Registry."), _("Preferences"), MB_ERR);
	return rc;
    }
    return 0;
}


int
set_reg_winpt_flag (const char *name, int val)
{
    return set_reg_entry (HKEY_CURRENT_USER, WINPT_REG, name, val? "1" : "0");
}


int
get_reg_winpt_flag (const char *name)
{
    char *buf;
    int flag = 0;

    buf = get_reg_entry (HKEY_CURRENT_USER, WINPT_REG, name);
    if (buf && buf[0] == '1')
	flag = 1;
    else if (!buf || (buf && buf[0] != '0'))
	flag = -1;
    free_if_alloc (buf);
    return flag;
}


/* Retrieve the winpt preferences from the registry. */
int
get_reg_winpt_prefs (winpt_prefs_t opt)
{
    char *val = NULL;
    size_t i;

    for (i=1; i < DIM (cfg); i++) {
	val = get_reg_entry (HKEY_CURRENT_USER, WINPT_REG, cfg[i]);
	if (!val || *val == ' ') {
	    free_if_alloc (val);
	    continue;
	}
	switch (i) {
	case CFG_CACHETIME:
	    opt->cache_time = atoi (val);
	    break;
	case CFG_WORDWRAP:
	    opt->word_wrap = atoi (val);
	    break;
	case CFG_FILEEXT:
	    opt->default_ext = atoi (val);
	    break;
	case CFG_NOZIP_MMEDIA:
	    opt->no_zip_mmedia = atoi (val);
	    break;
	case CFG_VIEWER:
	    opt->use_viewer = atoi (val);
	    break;
	case CFG_DISHOTKEYS:
	    opt->no_hotkeys = atoi (val);
	    break;
	case CFG_ALWAYSTRUST:
	    opt->always_trust = atoi (val);
	    break;
	case CFG_AUTOBACKUP:
	    opt->auto_backup = atoi (val);
	    break;
	case CFG_AUTOBAKMODE:
	    opt->backup.mode = atoi (val);
	    break;
	case CFG_EXPERT:
	    opt->expert = atoi (val);
	    break;
	case CFG_FM_PROGRESS:
	    opt->fm.progress = atoi (val);
	    break;

	case CFG_BACKUP_INC_SKR:
	    opt->backup.include_secr = atoi (val);
	    break;
	}
	free_if_alloc (val);
    }    

    val = get_reg_entry (HKEY_CURRENT_USER, WINPT_REG, "BackupPath");
    if (val && val[0] != ' ')
	opt->backup.path = m_strdup (val);
    free_if_alloc (val);
    
    for (i=0; wpt_hotkeys[i].name; i++) {
	wpt_hotkeys[i].enabled = 0;
	val = get_reg_entry (HKEY_CURRENT_USER, WINPT_REG, wpt_hotkeys[i].name);
	if (val && val[0] != ' ') {
	    wpt_hotkeys[i].key = *val;
	    wpt_hotkeys[i].enabled = 1;
	}
	free_if_alloc (val);
    }
    return 0;
}


char*
get_reg_entry_keyserver (const char *name)
{
    char *p = get_reg_entry (HKEY_CURRENT_USER, WINPT_REG"\\Keyserver", name);
    if (p && (!strcmp (p, "") || strlen (p) == 0)) {
	free_if_alloc (p);
	return NULL;
    }
    return p;
}


int
set_reg_entry_keyserver (const char *name, const char *val)
{
    return set_reg_entry (HKEY_CURRENT_USER, WINPT_REG"\\Keyserver", name, val);
}

